"""
This module provides a function to check the relative size of the corpus callosum

"""
# -----------------------------------------------------------------------------


def checkCCSize(subjects_dir, subject):
    """
    Check the relative size of the corpus callosum.

    This function evaluates the relative size of the corpus callosum by computing
    the sum of its volumes from the aseg.stats file and dividing it by the total
    intracranial volume.

    Parameters
    ----------
    subjects_dir : str
        The directory containing subject data.
    subject : str
        The name of the subject.

    Returns
    -------
    relative_cc : float
        The relative size of the corpus callosum.

    Notes
    -----
    It requires a valid stats/aseg.stats file.
    """
    # Imports
    import logging
    import os

    import numpy as np

    # Message
    logging.info("Checking size of the corpus callosum ...")

    # Get file name and read contents
    path_stats_file = os.path.join(subjects_dir, subject, "stats", "aseg.stats")

    with open(path_stats_file) as stats_file:
        aseg_stats = stats_file.read().splitlines()

    # Initialize
    cc_elements = [
        "CC_Posterior",
        "CC_Mid_Posterior",
        "CC_Central",
        "CC_Mid_Anterior",
        "CC_Anterior",
    ]

    sum_cc = 0.0

    intracranial_volume = None

    # Loop through the cc elements
    for cc_segmentation in cc_elements:
        # Look for the element in the aseg.stats line
        for aseg_stat_line in aseg_stats:
            # If the segmentation is found, compute the sum and return it.
            if cc_segmentation in aseg_stat_line:
                sum_cc += float(aseg_stat_line.split()[3])
            elif "# Measure EstimatedTotalIntraCranialVol" in aseg_stat_line:
                intracranial_volume = float(aseg_stat_line.split(",")[3])

    if sum_cc > 0.0 and intracranial_volume is not None:
        relative_cc = sum_cc / intracranial_volume
    else:
        relative_cc = np.nan

    logging.info(
        "Relative size of the corpus callosum is " + f"{relative_cc:.4}"
    )

    # Return
    return relative_cc



"""
This module provides a function to determine rotation angles of the Talairach transform

"""
# -------------------------------------------------------------------------------------


def checkRotation(subjects_dir, subject):
    """
    CheckRotation determines rotation angles of the Talairach transform.

    Parameters
    ----------
    subjects_dir : str
        The directory containing subject data.
    subject : str
        The name of the subject.

    Returns
    -------
    rot_tal_x : float
        Rotation angle around the x-axis in radians.
    rot_tal_y : float
        Rotation angle around the y-axis in radians.
    rot_tal_z : float
        Rotation angle around the z-axis in radians.

    Notes
    -----
    The unit of rotations is radians.

    Requires a valid mri/transforms/talairach.lta file and the 'transforms3d'
    python package. If not found, NaNs will be returned.
    """
    import importlib.util
    import logging
    import os
    import re
    import warnings

    import numpy as np

    # settings

    logging.captureWarnings(True)

    # message

    logging.info("Computing Talairach rotation angles ...")

    if importlib.util.find_spec("transforms3d") is None:
        warnings.warn(
            "WARNING: 'transforms3d' package required for running this script, returning NaNs.",
            stacklevel = 2
        )
        return np.nan, np.nan, np.nan
    else:
        import transforms3d as tr

    # read talairach.lta

    if not os.path.isfile(
        os.path.join(subjects_dir, subject, "mri", "transforms", "talairach.lta")
    ):
        warnings.warn(
            "WARNING: could not open "
            + os.path.join(subjects_dir, subject, "mri", "transforms", "talairach.lta")
            + ", returning NaNs.",
            stacklevel = 2
        )
        return np.nan, np.nan, np.nan

    with open(
        os.path.join(subjects_dir, subject, "mri", "transforms", "talairach.lta")
    ) as datafile:
        lines = datafile.readlines()

    # get first four rows with three entries in exp notation

    mat = list()
    for line in lines:
        res = re.search(
            "^[\\-0-9]+\\.[0-9]+e[\\-\\+][0-9]+ [\\-0-9]+\\.[0-9]+e[\\-\\+][0-9]+ [\\-0-9]+\\.[0-9]+e[\\-\\+][0-9]+ [\\-0-9]+\\.[0-9]+e[\\-\\+][0-9]+",
            line,
        )
        if res is not None:
            mat.append(
                [
                    float(x)
                    for x in res.group().replace("\n", "").replace(";", "").split()
                ]
            )
    mat = np.array(mat)

    # get translation, rotation, scale/zoom, and shear matrices

    T, R, Z, S = tr.affines.decompose44(mat)

    # now need to decompose R into euler angles; note this implies 'sxyz'
    # (=static/extrinsic, 1:x, 2:y, 3:z rotation type and sequence)
    # https://matthew-brett.github.io/transforms3d/reference/transforms3d.euler.html

    rot_x, rot_y, rot_z = tr.euler.mat2euler(R)

    logging.info(
        "Found Talairach rotation angles: x = "
        + f"{rot_x:.3}"
        + ", y = "
        + f"{rot_y:.3}"
        + ", z = "
        + f"{rot_z:.3}"
        + " radians.",
    )

    return rot_x, rot_y, rot_z



"""
fsqc CLI
"""

from ..fsqcMain import _parse_arguments, run_fsqc


def main():
    # say hello
    print("")
    print("-----------------------------")
    print("fsqc")
    print("-----------------------------")
    print("")

    # parse arguments
    argsDict = _parse_arguments()

    # run fsqc
    if argsDict is not None:
        run_fsqc(subjects_dir=None, output_dir=None, argsDict=argsDict)



"""
This module provides a function to compute the WM/GM contrast SNR.
"""


# -----------------------------------------------------------------------------
def checkContrast(subjects_dir, subject):
    """
    Compute the WM/GM contrast SNR.

    This function computes the WM/GM contrast SNR based on the output of the
    pctsurfcon function.

    Parameters
    ----------
    subjects_dir : str
        The directory containing subject data.
    subject : str
        The name of the subject.

    Returns
    -------
    con_lh_snr : float
        Signal-to-noise ratio (SNR) for the left hemisphere WM/GM contrast.
    con_rh_snr : float
        Signal-to-noise ratio (SNR) for the right hemisphere WM/GM contrast.

    Notes
    -----
    It requires surf/[lr]h.w-g.pct.mgh and label/[lr]h.cortex.label.
    If not found, NaNs will be returned.
    """
    # Imports
    import logging
    import os
    import warnings

    import nibabel
    import numpy

    from fsqc.fsqcUtils import importMGH

    # Settings
    logging.captureWarnings(True)

    # Message
    logging.info("Checking WM/GM contrast SNR ...")

    # Check if files exist
    path_pct_lh = os.path.join(subjects_dir, subject, "surf", "lh.w-g.pct.mgh")
    if not os.path.exists(path_pct_lh):
        warnings.warn("WARNING: could not find " + path_pct_lh + ", returning NaNs",
            stacklevel = 2)
        return numpy.nan

    path_pct_rh = os.path.join(subjects_dir, subject, "surf", "rh.w-g.pct.mgh")
    if not os.path.exists(path_pct_rh):
        warnings.warn("WARNING: could not find " + path_pct_rh + ", returning NaNs",
            stacklevel = 2)
        return numpy.nan

    path_label_cortex_lh = os.path.join(
        subjects_dir, subject, "label", "lh.cortex.label"
    )
    if not os.path.exists(path_label_cortex_lh):
        warnings.warn(
            "WARNING: could not find " + path_label_cortex_lh + ", returning NaNs",
            stacklevel = 2
        )
        return numpy.nan

    path_label_cortex_rh = os.path.join(
        subjects_dir, subject, "label", "rh.cortex.label"
    )
    if not os.path.exists(path_label_cortex_rh):
        warnings.warn(
            "WARNING: could not find " + path_label_cortex_rh + ", returning NaNs",
            stacklevel = 2
        )
        return numpy.nan

    # Get the data from the mgh files
    con_lh = importMGH(path_pct_lh)
    con_rh = importMGH(path_pct_rh)

    label_array_lh = nibabel.freesurfer.io.read_label(path_label_cortex_lh)
    label_array_rh = nibabel.freesurfer.io.read_label(path_label_cortex_rh)

    # Only take the values of the cortex to compute the contrast control
    con_lh = numpy.take(con_lh, label_array_lh)
    con_rh = numpy.take(con_rh, label_array_rh)

    # Compute the Contrast to noise ratio
    con_lh_mean = numpy.mean(con_lh)
    con_lh_std = numpy.std(con_lh)
    con_lh_snr = con_lh_mean / con_lh_std
    logging.info(
        "WM/GM contrast SNR for the left hemisphere: " + f"{con_lh_snr:.4}"
    )

    con_rh_mean = numpy.mean(con_rh)
    con_rh_std = numpy.std(con_rh)
    con_rh_snr = con_rh_mean / con_rh_std
    logging.info(
        "WM/GM contrast SNR for the right hemisphere: " + f"{con_rh_snr:.4}"
    )

    # Return
    return con_lh_snr, con_rh_snr



"""
This module provides a function to evaluate potential missegmentation of the fornix

"""
# -----------------------------------------------------------------------------


def evaluateFornixSegmentation(
    SUBJECT,
    SUBJECTS_DIR,
    OUTPUT_DIR,
    CREATE_SCREENSHOT=True,
    SCREENSHOTS_OUTFILE=None,
    RUN_SHAPEDNA=True,
    N_EIGEN=15,
    WRITE_EIGEN=True,
):
    """
    Evaluate potential missegmentation of the fornix.

    This script assesses the potential missegmentation of the fornix,
    which might erroneously be attached to the 'corpus callosum' label.

    It applies the cc_up.lta transform to the norm.mgz and aseg files,
    creating a binary corpus callosum mask and surface. The resulting
    files are saved to subject-specific directories within
    the 'fornix' subdirectory of the output directory.

    If the corresponding arguments are set to 'True', the script also
    creates screenshots and runs a shape analysis of the
    corpus callosum surface. Resulting files are saved to the same directory
    as indicated above.

    Parameters
    ----------
    SUBJECT : str
        The subject identifier.
    SUBJECTS_DIR : str
        The directory containing subject data.
    OUTPUT_DIR : str
        The output directory.
    CREATE_SCREENSHOT : bool, optional (default: True)
        Whether to create screenshots.
    SCREENSHOTS_OUTFILE : str or list, optional (default: None)
        File or list of files for screenshots.
    RUN_SHAPEDNA : bool, optional (default: True)
        Whether to run shape analysis.
    N_EIGEN : int, optional (default: 30)
        Number of Eigenvalues for shape analysis.
    WRITE_EIGEN : bool, optional (default: True)
        Write csv file with eigenvalues (or nans) to output directory.

    Returns
    -------
    numpy.ndarray
        A numpy array of N_EIGEN eigenvalues if RUN_SHAPEDNA is True,
        otherwise a numpy array of NaNs of the same dimension.
    """
    # imports

    import logging
    import os
    import warnings

    import nibabel as nb
    import numpy as np
    import pandas as pd

    from fsqc.createScreenshots import createScreenshots
    from fsqc.fsqcUtils import applyTransform, binarizeImage

    # --------------------------------------------------------------------------
    # check files

    logging.captureWarnings(True)

    if not os.path.isfile(
        os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "transforms", "cc_up.lta")
    ):
        warnings.warn(
            "WARNING: could not find "
            + os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "transforms", "cc_up.lta")
            + ", returning NaNs", stacklevel = 2
        )

        out = np.empty(N_EIGEN)
        out[:] = np.nan

        return out

    elif not os.path.isfile(os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "aseg.mgz")):
        warnings.warn(
            "WARNING: could not find "
            + os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "aseg.mgz")
            + ", returning NaNs", stacklevel = 2
        )

        out = np.empty(N_EIGEN)
        out[:] = np.nan

        return out

    elif not os.path.isfile(os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz")):
        warnings.warn(
            "WARNING: could not find "
            + os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz")
            + ", returning NaNs", stacklevel = 2
        )

        out = np.empty(N_EIGEN)
        out[:] = np.nan

        return out

    if SCREENSHOTS_OUTFILE is None:
        SCREENSHOTS_OUTFILE = os.path.join(OUTPUT_DIR, "cc.png")

    # --------------------------------------------------------------------------
    # conduct transform for aseg and norm

    applyTransform(
        os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "aseg.mgz"),
        os.path.join(OUTPUT_DIR, "asegCCup.mgz"),
        mat_file=os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "transforms", "cc_up.lta"),
        interp="nearest",
    )

    # when using 'make_upright', conducting the transform for norm.mgz is no
    # longer necessary (and will produce the same results)

    applyTransform(
        os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz"),
        os.path.join(OUTPUT_DIR, "normCCup.mgz"),
        mat_file=os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "transforms", "cc_up.lta"),
        interp="cubic",
    )

    # create fornix mask

    binarizeImage(
        os.path.join(OUTPUT_DIR, "asegCCup.mgz"),
        os.path.join(OUTPUT_DIR, "cc.mgz"),
        match=[251, 252, 253, 254, 255],
    )

    # --------------------------------------------------------------------------
    # create screenshot

    if CREATE_SCREENSHOT is True:
        hdr = nb.load(os.path.join(OUTPUT_DIR, "asegCCup.mgz"))
        x_coord = np.matmul(
            hdr.header.get_vox2ras_tkr(), np.array((128, 128, 128, 1))[:, np.newaxis]
        )[0]

        createScreenshots(
            SUBJECT=SUBJECT,
            SUBJECTS_DIR=SUBJECTS_DIR,
            INTERACTIVE=False,
            VIEWS=[("x", x_coord - 1), ("x", x_coord), ("x", x_coord + 1)],
            LAYOUT=(1, 3),
            BASE=os.path.join(OUTPUT_DIR, "normCCup.mgz"),
            OVERLAY=os.path.join(OUTPUT_DIR, "cc.mgz"),
            SURF=None,
            OUTFILE=SCREENSHOTS_OUTFILE,
        )

    # --------------------------------------------------------------------------
    # run shapeDNA

    if RUN_SHAPEDNA is True:
        import nibabel as nb
        from lapy import TriaMesh, shapedna

        surf = nb.freesurfer.io.read_geometry(
            os.path.join(OUTPUT_DIR, "cc.surf"), read_metadata=True
        )

        evDict = shapedna.compute_shapedna(TriaMesh(v=surf[0], t=surf[1]), k=N_EIGEN)

        ev = evDict["Eigenvalues"]
        evec = evDict["Eigenvectors"]

        d = dict()
        d["Refine"] = 0
        d["Degree"] = 1
        d["Dimension"] = 2
        d["Elements"] = len(surf[1])
        d["DoF"] = len(surf[0])
        d["NumEW"] = N_EIGEN
        d["Eigenvalues"] = ev
        d["Eigenvectors"] = evec

        # return
        out = d["Eigenvalues"]

    else:
        out = np.empty(N_EIGEN)
        out[:] = np.nan

    # write output
    if WRITE_EIGEN is True:
        pd.DataFrame(out).transpose().to_csv(
            os.path.join(OUTPUT_DIR, SUBJECT + ".fornix.csv"), na_rep="NA", index=False
        )

    # --------------------------------------------------------------------------
    # return

    return out



"""
This module provides a function to evaluate potential missegmentation of the hypothalamus

"""

# -----------------------------------------------------------------------------


def evaluateHypothalamicSegmentation(
    SUBJECT,
    SUBJECTS_DIR,
    OUTPUT_DIR,
    CREATE_SCREENSHOT=True,
    SCREENSHOTS_OUTFILE=None,
    SCREENSHOTS_ORIENTATION="radiological",
):
    """
    Evaluate potential missegmentation of the hypothalamus.

    This script evaluates potential missegmentation of the hypothalamus as
    created by FreeSurfer 7.2's module for hypothalamic segmentation:
    https://surfer.nmr.mgh.harvard.edu/fswiki/HypothalamicSubunits.

    If the corresponding arguments are set to 'True', the script will also
    create screenshots. Resulting files will be saved to the same directory
    as indicated above.

    Parameters
    ----------
    SUBJECT : str
        The subject identifier.
    SUBJECTS_DIR : str
        The directory containing subject data.
    OUTPUT_DIR : str
        The output directory.
    CREATE_SCREENSHOT : bool, optional, default: True
        Whether to create screenshots.
    SCREENSHOTS_OUTFILE : str or list, optional, default: None
        File or list of files for screenshots.
    SCREENSHOTS_ORIENTATION : str, optional, default: "radiological"
        Orientation for screenshots.

    Returns
    -------
    None
        This function return nothing.

    Notes
    -----
    Required Files:
    - mri/hypothalamic_subunits_seg.v1.mgz
    - mri/norm.mgz
    """
    # imports

    import logging
    import os

    import nibabel as nb
    import numpy as np
    from scipy import ndimage

    from fsqc.createScreenshots import createScreenshots
    from fsqc.fsqcUtils import binarizeImage

    # --------------------------------------------------------------------------
    # check files

    if not os.path.isfile(os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz")):
        logging.error(
            "ERROR: could not find "
            + os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz")
            + ", not running hypothalamus module."
        )

        raise ValueError("File not found")

    if not os.path.isfile(
        os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "hypothalamic_subunits_seg.v1.mgz")
    ):
        logging.error(
            "ERROR: could not find "
            + os.path.join(
                SUBJECTS_DIR, SUBJECT, "mri", "hypothalamic_subunits_seg.v1.mgz"
            )
            + ", not running hypothalamus module."
        )

        raise ValueError("File not found")

    if SCREENSHOTS_OUTFILE is None:
        SCREENSHOTS_OUTFILE = os.path.join(OUTPUT_DIR, "hypothalamus.png")

    # --------------------------------------------------------------------------
    # create mask

    binarizeImage(
        os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "hypothalamic_subunits_seg.v1.mgz"),
        os.path.join(OUTPUT_DIR, "hypothalamus.mgz"),
        match=[801, 802, 803, 804, 805, 806, 807, 808, 809, 810],
    )

    # --------------------------------------------------------------------------
    # get centroids

    seg = nb.load(
        os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "hypothalamic_subunits_seg.v1.mgz")
    )
    seg_data = seg.get_fdata()
    seg_labels = np.setdiff1d(np.unique(seg_data), 0)

    centroids = np.array(ndimage.center_of_mass(seg_data, seg_data, seg_labels))
    centroids = np.concatenate((seg_labels[:, np.newaxis], centroids), axis=1)

    vox2ras_tkr = seg.header.get_vox2ras_tkr()

    ctr_tkr = np.concatenate(
        (centroids[:, 1:4], np.ones((centroids.shape[0], 1))), axis=1
    )
    ctr_tkr = np.matmul(vox2ras_tkr, ctr_tkr.T).T
    ctr_tkr = np.concatenate(
        (np.array(centroids[:, 0], ndmin=2).T, ctr_tkr[:, 0:3]), axis=1
    )

    #

    ctr_tkr_x0 = ctr_tkr[3, 1]
    ctr_tkr_x1 = ctr_tkr[8, 1]
    ctr_tkr_y0 = (ctr_tkr[0, 2] + ctr_tkr[5, 2]) / 2
    ctr_tkr_y1 = (ctr_tkr[1, 2] + ctr_tkr[6, 2]) / 2
    ctr_tkr_y2 = (ctr_tkr[3, 2] + ctr_tkr[8, 2]) / 2
    ctr_tkr_y3 = (ctr_tkr[4, 2] + ctr_tkr[9, 2]) / 2
    ctr_tkr_y4 = (ctr_tkr[2, 2] + ctr_tkr[7, 2]) / 2
    ctr_tkr_z0 = (ctr_tkr[2, 3] + ctr_tkr[7, 3]) / 2
    ctr_tkr_z1 = (ctr_tkr[4, 3] + ctr_tkr[9, 3]) / 2

    # set ranges for cropping the image (assuming RAS coordinates)

    ctr_tkr_x0_xlim = [ctr_tkr[3, 2] - 20, ctr_tkr[3, 2] + 20]
    ctr_tkr_x1_xlim = [ctr_tkr[8, 2] - 20, ctr_tkr[8, 2] + 20]

    ctr_tkr_y0_xlim = [
        (ctr_tkr[0, 1] + ctr_tkr[5, 1]) / 2 - 20,
        (ctr_tkr[0, 1] + ctr_tkr[5, 1]) / 2 + 20,
    ]
    ctr_tkr_y1_xlim = [
        (ctr_tkr[1, 1] + ctr_tkr[6, 1]) / 2 - 20,
        (ctr_tkr[1, 1] + ctr_tkr[6, 1]) / 2 + 20,
    ]
    ctr_tkr_y2_xlim = [
        (ctr_tkr[3, 1] + ctr_tkr[8, 1]) / 2 - 20,
        (ctr_tkr[3, 1] + ctr_tkr[8, 1]) / 2 + 20,
    ]
    ctr_tkr_y3_xlim = [
        (ctr_tkr[4, 1] + ctr_tkr[9, 1]) / 2 - 20,
        (ctr_tkr[4, 1] + ctr_tkr[9, 1]) / 2 + 20,
    ]
    ctr_tkr_y4_xlim = [
        (ctr_tkr[2, 1] + ctr_tkr[7, 1]) / 2 - 20,
        (ctr_tkr[2, 1] + ctr_tkr[7, 1]) / 2 + 20,
    ]

    ctr_tkr_z0_xlim = [
        (ctr_tkr[2, 1] + ctr_tkr[7, 1]) / 2 - 20,
        (ctr_tkr[2, 1] + ctr_tkr[7, 1]) / 2 + 20,
    ]
    ctr_tkr_z1_xlim = [
        (ctr_tkr[4, 1] + ctr_tkr[9, 1]) / 2 - 20,
        (ctr_tkr[4, 1] + ctr_tkr[9, 1]) / 2 + 20,
    ]

    ctr_tkr_x0_ylim = [ctr_tkr[3, 3] - 20, ctr_tkr[3, 3] + 20]
    ctr_tkr_x1_ylim = [ctr_tkr[8, 3] - 20, ctr_tkr[8, 3] + 20]

    ctr_tkr_y0_ylim = [
        (ctr_tkr[0, 3] + ctr_tkr[5, 3]) / 2 - 20,
        (ctr_tkr[0, 3] + ctr_tkr[5, 3]) / 2 + 20,
    ]
    ctr_tkr_y1_ylim = [
        (ctr_tkr[1, 3] + ctr_tkr[6, 3]) / 2 - 20,
        (ctr_tkr[1, 3] + ctr_tkr[6, 3]) / 2 + 20,
    ]
    ctr_tkr_y2_ylim = [
        (ctr_tkr[3, 3] + ctr_tkr[8, 3]) / 2 - 20,
        (ctr_tkr[3, 3] + ctr_tkr[8, 3]) / 2 + 20,
    ]
    ctr_tkr_y3_ylim = [
        (ctr_tkr[4, 3] + ctr_tkr[9, 3]) / 2 - 20,
        (ctr_tkr[4, 3] + ctr_tkr[9, 3]) / 2 + 20,
    ]
    ctr_tkr_y4_ylim = [
        (ctr_tkr[2, 3] + ctr_tkr[7, 3]) / 2 - 20,
        (ctr_tkr[2, 3] + ctr_tkr[7, 3]) / 2 + 20,
    ]

    ctr_tkr_z0_ylim = [
        (ctr_tkr[2, 2] + ctr_tkr[7, 2]) / 2 - 20,
        (ctr_tkr[2, 2] + ctr_tkr[7, 2]) / 2 + 20,
    ]
    ctr_tkr_z1_ylim = [
        (ctr_tkr[4, 2] + ctr_tkr[9, 2]) / 2 - 20,
        (ctr_tkr[4, 2] + ctr_tkr[9, 2]) / 2 + 20,
    ]

    XLIM = [
        ctr_tkr_x0_xlim,
        ctr_tkr_x1_xlim,
        ctr_tkr_y0_xlim,
        ctr_tkr_y1_xlim,
        ctr_tkr_y2_xlim,
        ctr_tkr_y3_xlim,
        ctr_tkr_y4_xlim,
        ctr_tkr_z0_xlim,
        ctr_tkr_z1_xlim,
    ]

    YLIM = [
        ctr_tkr_x0_ylim,
        ctr_tkr_x1_ylim,
        ctr_tkr_y0_ylim,
        ctr_tkr_y1_ylim,
        ctr_tkr_y2_ylim,
        ctr_tkr_y3_ylim,
        ctr_tkr_y4_ylim,
        ctr_tkr_z0_ylim,
        ctr_tkr_z1_ylim,
    ]

    # --------------------------------------------------------------------------
    # create screenshot

    if CREATE_SCREENSHOT is True:
        createScreenshots(
            SUBJECT=SUBJECT,
            SUBJECTS_DIR=SUBJECTS_DIR,
            INTERACTIVE=False,
            VIEWS=[
                ("x", ctr_tkr_x0),
                ("x", ctr_tkr_x1),
                ("y", ctr_tkr_y0),
                ("y", ctr_tkr_y1),
                ("y", ctr_tkr_y2),
                ("y", ctr_tkr_y3),
                ("y", ctr_tkr_y4),
                ("z", ctr_tkr_z0),
                ("z", ctr_tkr_z1),
            ],
            LAYOUT=(1, 9),
            BASE=os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz"),
            OVERLAY=os.path.join(
                SUBJECTS_DIR, SUBJECT, "mri", "hypothalamic_subunits_seg.v1.mgz"
            ),
            SURF=None,
            OUTFILE=SCREENSHOTS_OUTFILE,
            ORIENTATION=SCREENSHOTS_ORIENTATION,
            XLIM=XLIM,
            YLIM=YLIM,
        )



import argparse

from .. import sys_info


def run():
    """Run sys_info() command."""
    parser = argparse.ArgumentParser(
        prog=f"{__package__.split('.')[0]}-sys_info", description="sys_info"
    )
    parser.add_argument(
        "--developer",
        help="display information for optional dependencies",
        action="store_true",
    )
    args = parser.parse_args()

    sys_info(developer=args.developer)



"""
This module provides a function to evaluate potential outliers in the
aseg.stats, aparc.stats, and hypothalamic and hippocampal values (if present).

"""

# ------------------------------------------------------------------------------
# subfunctions


def readAsegStats(path_aseg_stats):
    """
    Read FreeSurfer aseg.stats files.

    Parameters
    ----------
    path_aseg_stats : str
        Path to the aseg.stats file.

    Returns
    -------
    dict
        A dictionary containing FreeSurfer aseg measures.
    """
    # read file
    with open(path_aseg_stats) as stats_file:
        aseg_stats = stats_file.read().splitlines()

    # initialize
    aseg = dict()

    # read measures
    for line in aseg_stats:
        if "# Measure BrainSeg," in line:
            aseg.update({"aseg.BrainSeg": float(line.split(",")[3])})
        elif "# Measure BrainSegNotVent," in line:
            aseg.update({"aseg.BrainSegNotVent": float(line.split(",")[3])})
        elif "# Measure BrainSegNotVentSurf," in line:
            aseg.update({"aseg.BrainSegNotVentSurf": float(line.split(",")[3])})
        elif "# Measure VentricleChoroidVol," in line:
            aseg.update({"aseg.VentricleChoroidVol": float(line.split(",")[3])})
        elif "# Measure lhCortex," in line:
            aseg.update({"aseg.lhCortex": float(line.split(",")[3])})
        elif "# Measure rhCortex," in line:
            aseg.update({"aseg.rhCortex": float(line.split(",")[3])})
        elif "# Measure Cortex," in line:
            aseg.update({"aseg.Cortex": float(line.split(",")[3])})
        elif "# Measure lhCerebralWhiteMatter," in line:
            aseg.update({"aseg.lhCerebralWhiteMatter": float(line.split(",")[3])})
        elif "# Measure rhCerebralWhiteMatter," in line:
            aseg.update({"aseg.rhCerebralWhiteMatter": float(line.split(",")[3])})
        elif "# Measure CerebralWhiteMatter," in line:
            aseg.update({"aseg.CerebralWhiteMatter": float(line.split(",")[3])})
        elif "# Measure SubCortGray," in line:
            aseg.update({"aseg.SubCortGray": float(line.split(",")[3])})
        elif "# Measure TotalGray," in line:
            aseg.update({"aseg.TotalGray": float(line.split(",")[3])})
        elif "# Measure SupraTentorial," in line:
            aseg.update({"aseg.SupraTentorial": float(line.split(",")[3])})
        elif "# Measure SupraTentorialNotVent," in line:
            aseg.update({"aseg.SupraTentorialNotVent": float(line.split(",")[3])})
        elif "# Measure SupraTentorialNotVentVox," in line:
            aseg.update({"aseg.SupraTentorialNotVentVox": float(line.split(",")[3])})
        elif "# Measure Mask," in line:
            aseg.update({"aseg.Mask": float(line.split(",")[3])})
        elif "# Measure BrainSegVol-to-eTIV," in line:
            aseg.update({"aseg.BrainSegVol_to_eTIV": float(line.split(",")[3])})
        elif "# Measure MaskVol-to-eTIV," in line:
            aseg.update({"aseg.MaskVol_to_eTIV": float(line.split(",")[3])})
        elif "# Measure lhSurfaceHoles," in line:
            aseg.update({"aseg.lhSurfaceHoles": float(line.split(",")[3])})
        elif "# Measure rhSurfaceHoles," in line:
            aseg.update({"aseg.rhSurfaceHoles": float(line.split(",")[3])})
        elif "# Measure SurfaceHoles," in line:
            aseg.update({"aseg.SurfaceHoles": float(line.split(",")[3])})
        elif "# Measure EstimatedTotalIntraCranialVol," in line:
            aseg.update(
                {"aseg.EstimatedTotalIntraCranialVol": float(line.split(",")[3])}
            )
        elif "Left-Lateral-Ventricle" in line:
            aseg.update({"aseg.Left-Lateral-Ventricle": float(line.split()[3])})
        elif "Left-Inf-Lat-Vent" in line:
            aseg.update({"aseg.Left-Inf-Lat-Vent": float(line.split()[3])})
        elif "Left-Cerebellum-White-Matter" in line:
            aseg.update({"aseg.Left-Cerebellum-White-Matter": float(line.split()[3])})
        elif "Left-Cerebellum-Cortex" in line:
            aseg.update({"aseg.Left-Cerebellum-Cortex": float(line.split()[3])})
        elif "Left-Thalamus-Proper" in line:
            aseg.update({"aseg.Left-Thalamus-Proper": float(line.split()[3])})
        elif "Left-Caudate" in line:
            aseg.update({"aseg.Left-Caudate": float(line.split()[3])})
        elif "Left-Putamen" in line:
            aseg.update({"aseg.Left-Putamen": float(line.split()[3])})
        elif "Left-Pallidum" in line:
            aseg.update({"aseg.Left-Pallidum": float(line.split()[3])})
        elif "3rd-Ventricle" in line:
            aseg.update({"aseg.3rd-Ventricle": float(line.split()[3])})
        elif "4th-Ventricle" in line:
            aseg.update({"aseg.4th-Ventricle": float(line.split()[3])})
        elif "Brain-Stem" in line:
            aseg.update({"aseg.Brain-Stem": float(line.split()[3])})
        elif "Left-Hippocampus" in line:
            aseg.update({"aseg.Left-Hippocampus": float(line.split()[3])})
        elif "Left-Amygdala" in line:
            aseg.update({"aseg.Left-Amygdala": float(line.split()[3])})
        elif "CSF" in line:
            aseg.update({"aseg.CSF": float(line.split()[3])})
        elif "Left-Accumbens-area" in line:
            aseg.update({"aseg.Left-Accumbens-area": float(line.split()[3])})
        elif "Left-VentralDC" in line:
            aseg.update({"aseg.Left-VentralDC": float(line.split()[3])})
        elif "Left-vessel" in line:
            aseg.update({"aseg.Left-vessel": float(line.split()[3])})
        elif "Left-choroid-plexus" in line:
            aseg.update({"aseg.Left-choroid-plexus": float(line.split()[3])})
        elif "Right-Lateral-Ventricle" in line:
            aseg.update({"aseg.Right-Lateral-Ventricle": float(line.split()[3])})
        elif "Right-Inf-Lat-Vent" in line:
            aseg.update({"aseg.Right-Inf-Lat-Vent": float(line.split()[3])})
        elif "Right-Cerebellum-White-Matter" in line:
            aseg.update({"aseg.Right-Cerebellum-White-Matter": float(line.split()[3])})
        elif "Right-Cerebellum-Cortex" in line:
            aseg.update({"aseg.Right-Cerebellum-Cortex": float(line.split()[3])})
        elif "Right-Thalamus-Proper" in line:
            aseg.update({"aseg.Right-Thalamus-Proper": float(line.split()[3])})
        elif "Right-Caudate" in line:
            aseg.update({"aseg.Right-Caudate": float(line.split()[3])})
        elif "Right-Putamen" in line:
            aseg.update({"aseg.Right-Putamen": float(line.split()[3])})
        elif "Right-Pallidum" in line:
            aseg.update({"aseg.Right-Pallidum": float(line.split()[3])})
        elif "Right-Hippocampus" in line:
            aseg.update({"aseg.Right-Hippocampus": float(line.split()[3])})
        elif "Right-Amygdala" in line:
            aseg.update({"aseg.Right-Amygdala": float(line.split()[3])})
        elif "Right-Accumbens-area" in line:
            aseg.update({"aseg.Right-Accumbens-area": float(line.split()[3])})
        elif "Right-VentralDC" in line:
            aseg.update({"aseg.Right-VentralDC": float(line.split()[3])})
        elif "Right-vessel" in line:
            aseg.update({"aseg.Right-vessel": float(line.split()[3])})
        elif "Right-choroid-plexus" in line:
            aseg.update({"aseg.Right-choroid-plexus": float(line.split()[3])})
        elif "5th-Ventricle" in line:
            aseg.update({"aseg.5th-Ventricle": float(line.split()[3])})
        elif "WM-hypointensities" in line:
            aseg.update({"aseg.WM-hypointensities": float(line.split()[3])})
        elif "Left-WM-hypointensities" in line:
            aseg.update({"aseg.Left-WM-hypointensities": float(line.split()[3])})
        elif "Right-WM-hypointensities" in line:
            aseg.update({"aseg.Right-WM-hypointensities": float(line.split()[3])})
        elif "non-WM-hypointensities" in line:
            aseg.update({"aseg.non-WM-hypointensities": float(line.split()[3])})
        elif "Left-non-WM-hypointensities" in line:
            aseg.update({"aseg.Left-non-WM-hypointensities": float(line.split()[3])})
        elif "Right-non-WM-hypointensities" in line:
            aseg.update({"aseg.Right-non-WM-hypointensities": float(line.split()[3])})
        elif "Optic-Chiasm" in line:
            aseg.update({"aseg.Optic-Chiasm": float(line.split()[3])})
        elif "CC_Posterior" in line:
            aseg.update({"aseg.CC_Posterior": float(line.split()[3])})
        elif "CC_Mid_Posterior" in line:
            aseg.update({"aseg.CC_Mid_Posterior": float(line.split()[3])})
        elif "CC_Central" in line:
            aseg.update({"aseg.CC_Central": float(line.split()[3])})
        elif "CC_Mid_Anterior" in line:
            aseg.update({"aseg.CC_Mid_Anterior": float(line.split()[3])})
        elif "CC_Anterior" in line:
            aseg.update({"aseg.CC_Anterior": float(line.split()[3])})

    # return
    return aseg


def readAparcStats(path_aparc_stats, hemi):
    """
    Read FreeSurfer aparc.stats files.

    Parameters
    ----------
    path_aparc_stats : str
        Path to the aparc.stats file.
    hemi : str
        Hemisphere designation ('lh' for left, 'rh' for right).

    Returns
    -------
    tuple
        A tuple containing three dictionaries:

        - Header information for the aparc.stats file.
        - Detailed measures for different anatomical regions.
        - Thickness values for each region.
    """
    # read file
    with open(path_aparc_stats) as stats_file:
        aparc_stats = stats_file.read().splitlines()

    # read header
    header = dict()
    for line in aparc_stats:
        if "# Measure Cortex, NumVert, " in line:
            header.update(
                {
                    "aparc"
                    + "."
                    + hemi
                    + "."
                    + "CortexNumVert": float(line.split(",")[3])
                }
            )
        elif "# Measure Cortex, WhiteSurfArea," in line:
            header.update(
                {
                    "aparc"
                    + "."
                    + hemi
                    + "."
                    + "CortexWhiteSurfArea": float(line.split(",")[3])
                }
            )
        elif "# Measure Cortex, MeanThickness," in line:
            header.update(
                {
                    "aparc"
                    + "."
                    + hemi
                    + "."
                    + "CortexMeanThickness": float(line.split(",")[3])
                }
            )
        elif "# Measure BrainSeg," in line:
            header.update(
                {"aparc" + "." + hemi + "." + "BrainSeg": float(line.split(",")[3])}
            )
        elif "# Measure BrainSegNotVent," in line:
            header.update(
                {
                    "aparc"
                    + "."
                    + hemi
                    + "."
                    + "BrainSegNotVent": float(line.split(",")[3])
                }
            )
        elif "# Measure BrainSegNotVentSurf," in line:
            header.update(
                {
                    "aparc"
                    + "."
                    + hemi
                    + "."
                    + "BrainSegNotVentSurf": float(line.split(",")[3])
                }
            )
        elif "# Measure Cortex, CortexVol Total cortical gray matter volume," in line:
            # missing comma in fs60 and fs71, hence [2]
            header.update(
                {"aparc" + "." + hemi + "." + "Cortex": float(line.split(",")[2])}
            )
        elif "# Measure SupraTentorial," in line:
            header.update(
                {
                    "aparc"
                    + "."
                    + hemi
                    + "."
                    + "SupraTentorial": float(line.split(",")[3])
                }
            )
        elif "# Measure SupraTentorialNotVent," in line:
            header.update(
                {
                    "aparc"
                    + "."
                    + hemi
                    + "."
                    + "SupraTentorialNotVent": float(line.split(",")[3])
                }
            )
        elif "# Measure EstimatedTotalIntraCranialVol," in line:
            header.update(
                {
                    "aparc"
                    + "."
                    + hemi
                    + "."
                    + "EstimatedTotalIntraCranialVol": float(line.split(",")[3])
                }
            )

    # get column headers
    col_headers = None
    for line in aparc_stats:
        if "# ColHeaders " in line:
            col_headers = line.replace("# ColHeaders", "").split()

    # read aparc
    aparc = dict()
    if col_headers is not None:
        for line in aparc_stats:
            if "bankssts" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "bankssts": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "caudalanteriorcingulate" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "caudalanteriorcingulate": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "caudalmiddlefrontal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "caudalmiddlefrontal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "cuneus" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "cuneus": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "entorhinal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "entorhinal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "fusiform" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "fusiform": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "inferiorparietal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "inferiorparietal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "inferiortemporal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "inferiortemporal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "isthmuscingulate" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "isthmuscingulate": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "lateraloccipital" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "lateraloccipital": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "lateralorbitofrontal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "lateralorbitofrontal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "lingual" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "lingual": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "medialorbitofrontal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "medialorbitofrontal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "middletemporal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "middletemporal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "parahippocampal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "parahippocampal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "paracentral" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "paracentral": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "parsopercularis" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "parsopercularis": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "parsorbitalis" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "parsorbitalis": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "parstriangularis" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "parstriangularis": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "pericalcarine" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "pericalcarine": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "postcentral" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "postcentral": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "posteriorcingulate" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "posteriorcingulate": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "precentral" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "precentral": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "precuneus" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "precuneus": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "rostralanteriorcingulate" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "rostralanteriorcingulate": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "rostralmiddlefrontal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "rostralmiddlefrontal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "superiorfrontal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "superiorfrontal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "superiorparietal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "superiorparietal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "superiortemporal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "superiortemporal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "supramarginal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "supramarginal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "frontalpole" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "frontalpole": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "temporalpole" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "temporalpole": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "transversetemporal" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "transversetemporal": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )
            elif "insula" in line:
                aparc.update(
                    {
                        "aparc"
                        + "."
                        + hemi
                        + "."
                        + "insula": dict(
                            zip(col_headers[1:], [float(x) for x in line.split()[1:]])
                        )
                    }
                )

    # extract thickness values
    if col_headers is not None:
        thickness = dict()
        for x in aparc.keys():
            thickness[x] = aparc[x]["ThickAvg"]

    # return
    return header, aparc, thickness


def readHypothalamusStats(path_hypothalamus_stats):
    """
    Read hypothalamic volume files.

    Parameters
    ----------
    path_hypothalamus_stats : str
        Path to the hypothalamic volume file.

    Returns
    -------
    dict
        A dictionary containing hypothalamic volume information.
    """
    # imports
    import pandas as pd

    # read file
    hypothalamus_stats = pd.read_csv(path_hypothalamus_stats)

    #
    hypothalamus_stats.columns = "hypothalamus" + "." + hypothalamus_stats.columns

    #
    hypo = hypothalamus_stats.to_dict(orient="index")[0]

    #
    hypo.pop("hypothalamus.subject", None)

    # return
    return hypo


def readHippocampusStats(path_hippocampus_stats, hemi, prefix):
    """
    Read hippocampal or amygdala volume files.

    Parameters
    ----------
    path_hippocampus_stats : str
        Path to the volume file.
    hemi : str
        Hemisphere identifier (e.g., 'lh' or 'rh').
    prefix : str
        Prefix for column names.

    Returns
    -------
    dict
        Dictionary containing volume values with column names.
    """
    # imports
    import pandas as pd

    # read file
    hippocampus_stats = pd.read_csv(
        path_hippocampus_stats, sep=" ", header=None, names=["volume"], index_col=0
    )

    #
    hippo = hippocampus_stats.transpose()

    hippo.columns = prefix + "." + hemi + "." + hippo.columns

    hippo = hippo.to_dict(orient="index")["volume"]

    # return
    return hippo


# ------------------------------------------------------------------------------
# outlier table


def outlierTable():
    """
    Provide upper and lower bounds for volumes of several brain structures.

    Returns
    -------
    dict
        A dictionary containing upper and lower bounds for several brain structures.
    """
    # define

    outlierDict = dict(
        [
            (
                "Left-Accumbens-area",
                dict([("lower", 210.87844594754), ("upper", 718.01022026916)]),
            ),
            (
                "Right-Accumbens-area",
                dict([("lower", 304.86134907845), ("upper", 751.63838456345)]),
            ),
            (
                "Left-Amygdala",
                dict([("lower", 1179.73655974083), ("upper", 1935.09415214717)]),
            ),
            (
                "Right-Amygdala",
                dict([("lower", 1161.54746836742), ("upper", 2002.14187676668)]),
            ),
            (
                "Brain-Stem",
                dict([("lower", 18048.54263155760), ("upper", 25300.51090318110)]),
            ),
            (
                "Left-Caudate",
                dict([("lower", 2702.73311142764), ("upper", 4380.54479618196)]),
            ),
            (
                "Right-Caudate",
                dict([("lower", 2569.61140834210), ("upper", 4412.61035536070)]),
            ),
            (
                "Left-Hippocampus",
                dict([("lower", 3432.26483953083), ("upper", 4934.43236139507)]),
            ),
            (
                "Right-Hippocampus",
                dict([("lower", 3580.74371035841), ("upper", 5067.49668145829)]),
            ),
            (
                "Left-Pallidum",
                dict([("lower", 935.47686324176), ("upper", 1849.42861796994)]),
            ),
            (
                "Right-Pallidum",
                dict([("lower", 1078.14975428593), ("upper", 1864.08951102817)]),
            ),
            (
                "Left-Putamen",
                dict([("lower", 3956.23134409153), ("upper", 6561.97642872937)]),
            ),
            (
                "Right-Putamen",
                dict([("lower", 3768.88684356957), ("upper", 6142.52870810603)]),
            ),
            (
                "Left-Thalamus-Proper",
                dict([("lower", 6483.36121320953), ("upper", 9489.46749012527)]),
            ),
            (
                "Right-Thalamus-Proper",
                dict([("lower", 6065.70220487045), ("upper", 8346.88382091555)]),
            ),
            (
                "Left-VentralDC",
                dict([("lower", 3182.42264293449), ("upper", 4495.77412707751)]),
            ),
            (
                "Right-VentralDC",
                dict([("lower", 3143.88280953869), ("upper", 4407.63641978371)]),
            ),
        ]
    )

    # return
    return outlierDict


# ------------------------------------------------------------------------------
# main function


def outlierDetection(
    subjects,
    subjects_dir,
    output_dir,
    outlierDict,
    min_no_subjects=10,
    hypothalamus=False,
    hippocampus=False,
    hippocampus_label=None,
    fastsurfer=False,
):
    """
    Evaluate outliers in aseg.stats, [lr]h.aparc, and optional hypothalamic/hippocampal values.

    Parameters
    ----------
    subjects : list
        List of subject IDs.
    subjects_dir : str
        Path to the FreeSurfer subjects directory.
    output_dir : str
        Path to the output directory for saving results.
    outlierDict : dict
        Dictionary containing outlier thresholds for different measures.
    min_no_subjects : int, optional
        Minimum number of subjects required for analysis.
    hypothalamus : bool, optional
        Flag to include hypothalamic values in the analysis.
    hippocampus : bool, optional
        Flag to include hippocampal values in the analysis.
    hippocampus_label : str or None, optional
        Label to identify the hippocampus (e.g., "Hippocampus").
    fastsurfer : bool, optional
        Flag to use FastSurfer instead of FreeSurfer output files.

    Returns
    -------
    tuple
        A tuple containing three dictionaries:

        - outlierSampleNonparNum
        - outlierSampleParamNum
        - outlierNormsNum
    """
    # imports

    import csv
    import os

    import numpy as np
    import pandas as pd

    # create a dictionary with all data from all subjects, and create a list of all available keys

    regions = dict()

    all_regions_keys = list()

    for subject in subjects:
        # aseg
        path_aseg_stats = os.path.join(subjects_dir, subject, "stats", "aseg.stats")
        aseg_stats = readAsegStats(path_aseg_stats)
        regions[subject] = aseg_stats.copy()
        all_regions_keys.extend(list(aseg_stats.keys()))

        # aparc
        if fastsurfer is True:
            path_aparc_stats = os.path.join(
                subjects_dir, subject, "stats", "lh.aparc.DKTatlas.mapped.stats"
            )
        else:
            path_aparc_stats = os.path.join(
                subjects_dir, subject, "stats", "lh.aparc.stats"
            )
        aparc_header, aparc_stats, aparc_thickness = readAparcStats(
            path_aparc_stats, hemi="lh"
        )
        regions[subject].update(aparc_thickness)
        all_regions_keys.extend(list(aparc_thickness.keys()))

        if fastsurfer is True:
            path_aparc_stats = os.path.join(
                subjects_dir, subject, "stats", "rh.aparc.DKTatlas.mapped.stats"
            )
        else:
            path_aparc_stats = os.path.join(
                subjects_dir, subject, "stats", "rh.aparc.stats"
            )
        aparc_header, aparc_stats, aparc_thickness = readAparcStats(
            path_aparc_stats, hemi="rh"
        )
        regions[subject].update(aparc_thickness)
        all_regions_keys.extend(list(aparc_thickness.keys()))

        # hypothalamus

        if hypothalamus is True:
            path_hypothalamus_stats = os.path.join(
                subjects_dir, subject, "mri", "hypothalamic_subunits_volumes.v1.csv"
            )
            if os.path.exists(path_hypothalamus_stats):
                hypothalamus_stats = readHypothalamusStats(path_hypothalamus_stats)
                regions[subject].update(hypothalamus_stats)
                all_regions_keys.extend(list(hypothalamus_stats.keys()))

        # hippocampus + amygdala

        if hippocampus is True and hippocampus_label is not None:
            path_hippocampus_stats = os.path.join(
                subjects_dir,
                subject,
                "mri",
                "lh.hippoSfVolumes-" + hippocampus_label + ".txt",
            )
            if os.path.exists(path_hippocampus_stats):
                hippocampus_stats = readHippocampusStats(
                    path_hippocampus_stats, hemi="lh", prefix="hippocampus"
                )
                regions[subject].update(hippocampus_stats)
                all_regions_keys.extend(list(hippocampus_stats.keys()))

            path_hippocampus_stats = os.path.join(
                subjects_dir,
                subject,
                "mri",
                "rh.hippoSfVolumes-" + hippocampus_label + ".txt",
            )
            if os.path.exists(path_hippocampus_stats):
                hippocampus_stats = readHippocampusStats(
                    path_hippocampus_stats, hemi="rh", prefix="hippocampus"
                )
                regions[subject].update(hippocampus_stats)
                all_regions_keys.extend(list(hippocampus_stats.keys()))

            path_hippocampus_stats = os.path.join(
                subjects_dir,
                subject,
                "mri",
                "lh.amygNucVolumes-" + hippocampus_label + ".txt",
            )
            if os.path.exists(path_hippocampus_stats):
                hippocampus_stats = readHippocampusStats(
                    path_hippocampus_stats, hemi="lh", prefix="amygdala"
                )
                regions[subject].update(hippocampus_stats)
                all_regions_keys.extend(list(hippocampus_stats.keys()))

            path_hippocampus_stats = os.path.join(
                subjects_dir,
                subject,
                "mri",
                "rh.amygNucVolumes-" + hippocampus_label + ".txt",
            )
            if os.path.exists(path_hippocampus_stats):
                hippocampus_stats = readHippocampusStats(
                    path_hippocampus_stats, hemi="rh", prefix="amygdala"
                )
                regions[subject].update(hippocampus_stats)
                all_regions_keys.extend(list(hippocampus_stats.keys()))

    # sort keys

    all_regions_keys = set(all_regions_keys)

    all_regions_keys = (
        sorted(list(filter(lambda x: "aseg." in x, list(all_regions_keys))))
        + sorted(list(filter(lambda x: "aparc." in x, list(all_regions_keys))))
        + sorted(list(filter(lambda x: "hippocampus." in x, list(all_regions_keys))))
        + sorted(list(filter(lambda x: "amygdala." in x, list(all_regions_keys))))
        + sorted(list(filter(lambda x: "hypothalamus." in x, list(all_regions_keys))))
    )

    # compare individual data against sample statistics (if more than min_no_subjects cases)

    outlierSampleNonpar = dict()
    outlierSampleParam = dict()

    outlierSampleNonparNum = dict()
    outlierSampleParamNum = dict()

    if len(subjects) >= min_no_subjects:
        # compute means, sd, medians, and quantiles based on sample

        df = pd.DataFrame.from_dict(regions).transpose()

        iqr = np.percentile(df, 75, axis=0) - np.percentile(df, 25, axis=0)

        sample_nonpar_lower = dict(
            zip(df.columns, np.percentile(df, 25, axis=0) - 1.5 * iqr)
        )
        sample_nonpar_upper = dict(
            zip(df.columns, np.percentile(df, 75, axis=0) + 1.5 * iqr)
        )

        sample_param_lower = dict(np.mean(df, axis=0) - 2 * np.std(df, axis=0))
        sample_param_upper = dict(np.mean(df, axis=0) + 2 * np.std(df, axis=0))

        # compare individual data against sample statistics

        for subject in regions:
            nonparDict = dict()
            paramDict = dict()

            for key in regions[subject]:
                if (regions[subject][key] < sample_nonpar_lower[key]) or (
                    regions[subject][key] > sample_nonpar_upper[key]
                ):
                    nonparDict.update({key: True})
                else:
                    nonparDict.update({key: False})

                if (regions[subject][key] < sample_param_lower[key]) or (
                    regions[subject][key] > sample_param_upper[key]
                ):
                    paramDict.update({key: True})
                else:
                    paramDict.update({key: False})

            outlierSampleNonpar.update({subject: nonparDict})
            outlierSampleParam.update({subject: paramDict})

            outlierSampleNonparNum.update({subject: np.sum(list(nonparDict.values()))})
            outlierSampleParamNum.update({subject: np.sum(list(paramDict.values()))})

    else:
        for subject in regions:
            nonparDict = dict()
            paramDict = dict()

            for key in regions[subject]:
                nonparDict.update({key: np.nan})
                paramDict.update({key: np.nan})

            outlierSampleNonpar.update({subject: nonparDict})
            outlierSampleParam.update({subject: paramDict})

            outlierSampleNonparNum.update({subject: np.nan})
            outlierSampleParamNum.update({subject: np.nan})

    # compare individual data against normative values

    outlierNorms = dict()

    outlierNormsNum = dict()

    for subject in regions:
        normsDict = dict()

        for key in regions[subject]:
            # no prefixes in outlier table
            if key.startswith("aseg."):
                outlierKey = key.replace("aseg.", "")
            elif key.startswith("aparc."):
                outlierKey = key.replace("aparc.", "")
            elif key.startswith("hippocampus."):
                outlierKey = key.replace("hippocampus.", "")
            elif key.startswith("amygdala."):
                outlierKey = key.replace("amygdala.", "")
            elif key.startswith("hypothalamus."):
                outlierKey = key.replace("aseg.", "")

            if outlierKey in outlierDict:
                if (regions[subject][key] < outlierDict[outlierKey]["lower"]) or (
                    regions[subject][key] > outlierDict[outlierKey]["upper"]
                ):
                    normsDict.update({key: True})
                else:
                    normsDict.update({key: False})

            else:
                normsDict.update({key: np.nan})

        outlierNorms.update({subject: normsDict})

        outlierNormsNum.update({subject: np.nansum(list(normsDict.values()))})

    # write to csv files

    regionsFieldnames = ["subject"]
    regionsFieldnames.extend(all_regions_keys)

    with open(os.path.join(output_dir, "all.regions.stats"), "w") as datafile:
        csvwriter = csv.DictWriter(
            datafile,
            fieldnames=regionsFieldnames,
            delimiter=",",
            quotechar='"',
            quoting=csv.QUOTE_MINIMAL,
        )
        csvwriter.writeheader()
        for subject in sorted(list(regions.keys())):
            tmp = regions[subject]
            tmp.update({"subject": subject})
            csvwriter.writerow(tmp)

    with open(
        os.path.join(output_dir, "all.outliers.sample.nonpar.stats"), "w"
    ) as datafile:
        csvwriter = csv.DictWriter(
            datafile,
            fieldnames=regionsFieldnames,
            delimiter=",",
            quotechar='"',
            quoting=csv.QUOTE_MINIMAL,
        )
        csvwriter.writeheader()
        for subject in sorted(list(outlierSampleNonpar.keys())):
            tmp = outlierSampleNonpar[subject]
            tmp.update({"subject": subject})
            csvwriter.writerow(tmp)

    with open(
        os.path.join(output_dir, "all.outliers.sample.param.stats"), "w"
    ) as datafile:
        csvwriter = csv.DictWriter(
            datafile,
            fieldnames=regionsFieldnames,
            delimiter=",",
            quotechar='"',
            quoting=csv.QUOTE_MINIMAL,
        )
        csvwriter.writeheader()
        for subject in sorted(list(outlierSampleParam.keys())):
            tmp = outlierSampleParam[subject]
            tmp.update({"subject": subject})
            csvwriter.writerow(tmp)

    with open(os.path.join(output_dir, "all.outliers.norms.stats"), "w") as datafile:
        csvwriter = csv.DictWriter(
            datafile,
            fieldnames=regionsFieldnames,
            delimiter=",",
            quotechar='"',
            quoting=csv.QUOTE_MINIMAL,
        )
        csvwriter.writeheader()
        for subject in sorted(list(outlierNorms.keys())):
            tmp = outlierNorms[subject]
            tmp.update({"subject": subject})
            csvwriter.writerow(tmp)

    # return

    return outlierSampleNonparNum, outlierSampleParamNum, outlierNormsNum



"""
This module provides a function to create screenshots

"""

# -----------------------------------------------------------------------------


def createScreenshots(
    SUBJECT,
    SUBJECTS_DIR,
    OUTFILE,
    INTERACTIVE=True,
    LAYOUT="default",
    BASE="default",
    OVERLAY="default",
    LABELS=None,
    SURF="default",
    SURFCOLOR="default",
    VIEWS="default",
    XLIM=None,
    YLIM=None,
    BINARIZE=False,
    ORIENTATION="radiological",
):
    """
    Function to create screenshots.

    Parameters
    ----------
    SUBJECT : str
        The subject.
    SUBJECTS_DIR : str
        The subjects directory.
    OUTFILE : str
        The output file path.
    INTERACTIVE : bool, optional
        Flag for interactive mode, default is True.
    LAYOUT : list, optional
        The layout, default is "default".
    BASE : str, optional
        The base, default is "default".
        Load norm.mgz as default.
    OVERLAY : str, optional
        The overlay, default is "default".
        Load aseg.mgz as default.
        Can be None.
    LABELS : None or str, optional
        The labels, default is None.
    SURF : list, optional
        The surface, default is "default".
        Can be None.
    SURFCOLOR : list, optional
        The surface color, default is "default".
        Can be None.
    VIEWS : list, optional
        The views, default is "default".
    XLIM : None or list, optional
        The x limits, default is None.
    YLIM : None or list, optional
        The y limits, default is None.
    BINARIZE : bool, optional
        Flag for binarization, default is False.
    ORIENTATION : str, optional
        The orientation, default is "radiological".

    Notes
    -----
    LAYOUT, VIEWS can be lists or "default".

    SURF, SURFCOLOR can be lists, None, or "default".

    XLIM, YLIM can be lists of list two-element numeric lists or None; if given,
    length must match length of VIEWS. x and y refer to final image dimensions,
    not MR volume dimensions.
    """
    # -----------------------------------------------------------------------------
    # auxiliary functions

    def computeLayout(n):
        import numpy as np

        y = np.ceil(np.sqrt(n))

        x = y - np.divmod(y**2 - n, y)[0]

        return int(x), int(y)

    # -----------------------------------------------------------------------------
    # imports

    import logging
    import os
    import warnings

    import matplotlib
    import nibabel as nb
    import numpy as np

    if not INTERACTIVE:
        matplotlib.use("Agg")

    from matplotlib import pyplot as plt

    from fsqc.fsqcUtils import levelsetsTria, returnFreeSurferColorLUT

    # -----------------------------------------------------------------------------
    # settings

    logging.captureWarnings(True)

    FIGSIZE = 8

    FIGDPI = 100

    ALPHA = 0.5

    tol = 1e-16

    # -----------------------------------------------------------------------------
    # import image data

    if BASE == "default":
        norm = nb.load(os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz"))
    else:
        norm = nb.load(BASE)

    if OVERLAY is None:
        aseg = None
    elif OVERLAY == "default":
        aseg = nb.load(os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "aseg.mgz"))
    else:
        aseg = nb.load(OVERLAY)

    # -----------------------------------------------------------------------------
    # import surface data

    if SURF == "default":
        surflist = [
            os.path.join(SUBJECTS_DIR, SUBJECT, "surf", "lh.white"),
            os.path.join(SUBJECTS_DIR, SUBJECT, "surf", "rh.white"),
            os.path.join(SUBJECTS_DIR, SUBJECT, "surf", "lh.pial"),
            os.path.join(SUBJECTS_DIR, SUBJECT, "surf", "rh.pial"),
        ]
    else:
        surflist = SURF

    if SURF is None:
        surfcolor = None
    elif SURFCOLOR == "default" and SURF == "default":
        surfcolor = ["yellow", "yellow", "red", "red"]
    elif SURFCOLOR == "default" and SURF != "default":
        surfcolor = ["yellow"] * len(surflist)
    else:
        surfcolor = SURFCOLOR

    surf = list()

    if surflist is not None and surfcolor is not None:
        for i in range(len(surflist)):
            surf.append(nb.freesurfer.io.read_geometry(surflist[i], read_metadata=True))

    # -----------------------------------------------------------------------------
    # import colortable, compute auxiliary variables, and transform to matplotlib
    # colortable

    lut = returnFreeSurferColorLUT()

    # some fs7 labels are not present in fs6 LUT, check and add if necessary

    if not (
        np.isin(
            list(range(231, 247)) + [801, 802, 803, 804, 805, 806, 807, 808, 809, 810],
            lut[:, 0],
        ).all()
    ):
        lutAdd = np.array(
            (
                [801, "L_hypothalamus_anterior_inferior", 250, 255, 50, 0],
                [802, "L_hypothalamus_anterior_superior", 80, 200, 255, 0],
                [803, "L_hypothalamus_posterior", 255, 160, 0, 0],
                [804, "L_hypothalamus_tubular_inferior", 255, 160, 200, 0],
                [805, "L_hypothalamus_tubular_superior", 20, 180, 130, 0],
                [806, "R_hypothalamus_anterior_inferior", 250, 255, 50, 0],
                [807, "R_hypothalamus_anterior_superior", 80, 200, 255, 0],
                [808, "R_hypothalamus_posterior", 255, 160, 0, 0],
                [809, "R_hypothalamus_tubular_inferior", 255, 160, 200, 0],
                [810, "R_hypothalamus_tubular_superior", 20, 180, 130, 0],
                [231, "HP_body", 0, 255, 0, 0],
                [232, "HP_head", 255, 0, 0, 0],
                [233, "presubiculum-head", 32, 0, 32, 0],
                [234, "presubiculum-body", 64, 0, 64, 0],
                [235, "subiculum-head", 0, 0, 175, 0],
                [236, "subiculum-body", 0, 0, 255, 0],
                [237, "CA1-head", 175, 75, 75, 0],
                [238, "CA1-body", 255, 0, 0, 0],
                [239, "CA3-head", 0, 80, 0, 0],
                [240, "CA3-body", 0, 128, 0, 0],
                [241, "CA4-head", 120, 90, 50, 0],
                [242, "CA4-body", 196, 160, 128, 0],
                [243, "GC-ML-DG-head", 75, 125, 175, 0],
                [244, "GC-ML-DG-body", 32, 200, 255, 0],
                [245, "molecular_layer_HP-head", 100, 25, 25, 0],
                [246, "molecular_layer_HP-body", 128, 0, 0, 0],
            ),
            dtype=object,
        )

        lut = np.concatenate((lut, lutAdd), axis=0)

    lutEnum = dict(zip(lut[:, 0], range(len(lut[:, 0]))))

    lutTab = np.array(lut[:, (2, 3, 4, 5)] / 255, dtype="float32")
    lutTab[:, 3] = 1

    lutMap = matplotlib.colors.ListedColormap(lutTab)

    # -----------------------------------------------------------------------------
    # determine VIEWS

    if VIEWS == "default":
        CutsRRAS = [("x", -10), ("x", 10), ("y", 0), ("z", 0)]
    else:
        CutsRRAS = VIEWS

    # -----------------------------------------------------------------------------
    # check if the chosen VIEWS are feasible. If not feasible changing to the nearest feasible values

    m = norm.header.get_vox2ras_tkr()
    n = norm.header.get_data_shape()

    xyzIdx = np.array(
        np.meshgrid(
            np.linspace(0, n[0] - 1, n[0]),
            np.linspace(0, n[1] - 1, n[1]),
            np.linspace(0, n[2] - 1, n[2]),
        )
    )
    xyzIdxFlat3 = np.reshape(xyzIdx, (3, np.prod(n))).transpose()
    xyzIdxFlat4 = np.hstack((xyzIdxFlat3, np.ones((np.prod(n), 1))))
    rasIdxFlat3 = np.matmul(m, xyzIdxFlat4.transpose()).transpose()[:, 0:3]

    for i, icr in enumerate(CutsRRAS):
        if icr[0] == "x":
            iDim = 0
        elif icr[0] == "y":
            iDim = 1
        elif icr[0] == "z":
            iDim = 2

        if not np.any(rasIdxFlat3[:, iDim] == icr[1]):
            closestCutValue = rasIdxFlat3[
                np.abs(rasIdxFlat3[:, iDim] - icr[1]).argmin(), iDim
            ]
            logging.info(
                f"INFO: the VIEW {icr} will be changed to ('{icr[0]}', {closestCutValue:.2f}) so it is not"
                " necessary to interpolate volumetric data"
            )
            CutsRRAS[i] = (icr[0], closestCutValue)

    # -----------------------------------------------------------------------------
    # compute levelsets

    LVL = list()

    # will not run if surf is empty (intended)
    for s in range(len(surf)):
        sLVL = list()

        for i in range(len(CutsRRAS)):
            # determine dimension
            if CutsRRAS[i][0] == "x":
                iDim = 0
            elif CutsRRAS[i][0] == "y":
                iDim = 1
            elif CutsRRAS[i][0] == "z":
                iDim = 2

            # compute levelsets: e.g., LVL[SURF][VIEWS][vLVL|tLVL|iLVL][0][elementDim1][elementDim2]
            sLVL.append(
                levelsetsTria(
                    surf[s][0], surf[s][1], surf[s][0][:, iDim], CutsRRAS[i][1]
                )
            )

        LVL.append(sLVL)

    # -----------------------------------------------------------------------------
    # get data for norm

    normData = norm.get_fdata()

    normVals = normData

    # -----------------------------------------------------------------------------
    # get data for aseg and change to enumerated aseg so that it can be used as
    # index to lutMap

    if aseg is not None:
        asegData = aseg.get_fdata()

        if LABELS is not None:
            asegData = asegData * np.isin(asegData, LABELS)

        if BINARIZE is True:
            asegData = (asegData > 0).astype(int)

        asegUnique, asegIdx = np.unique(asegData, return_inverse=True)

        asegEnum = np.array([lutEnum[x] for x in asegUnique])

        asegVals = np.reshape(asegEnum[asegIdx], (aseg.shape))

    # -----------------------------------------------------------------------------
    # compile image data for plotting

    # note that we are using the TkReg RAS system, not RAS per se
    # this is because the surfaces are in TkReg RAS also

    # one considerable advantage of the (TkReg) RAS system is that it is solely
    # based on switching dimensions, not rotating at odd angles etc, so no need
    # for interpolating.

    # x_R y_R z_R c_R
    # x_A y_A z_A c_A
    # x_S y_S z_S c_S
    #   0   0   0   1

    # m = norm.header.get_vox2ras_tkr()
    # n = norm.header.get_data_shape()

    # xyzIdx = np.array(np.meshgrid(np.linspace(0, n[0] - 1, n[0]), np.linspace(0, n[1] - 1, n[1]), np.linspace(0, n[2] - 1, n[2])))
    # xyzIdxFlat3 = np.reshape(xyzIdx, (3, np.prod(n))).transpose()
    # xyzIdxFlat4 = np.hstack((xyzIdxFlat3, np.ones((np.prod(n), 1))))
    # rasIdxFlat3 = np.matmul(m, xyzIdxFlat4.transpose()).transpose()[:, 0:3]

    normValsRAS = list()
    if aseg is not None:
        asegValsRAS = list()

    for i in range(len(CutsRRAS)):
        # determine dimension
        if CutsRRAS[i][0] == "x":
            iDim = 0
        elif CutsRRAS[i][0] == "y":
            iDim = 1
        elif CutsRRAS[i][0] == "z":
            iDim = 2

        sel = np.array(
            xyzIdxFlat3[rasIdxFlat3[:, iDim] == CutsRRAS[i][1], :], dtype="int"
        )
        normValsRAS.append(
            np.squeeze(
                normVals[
                    np.min(sel[:, 0]) : np.max(sel[:, 0]) + 1,
                    np.min(sel[:, 1]) : np.max(sel[:, 1]) + 1,
                    np.min(sel[:, 2]) : np.max(sel[:, 2]) + 1,
                ]
            )
        )
        if aseg is not None:
            asegValsRAS.append(
                np.squeeze(
                    asegVals[
                        np.min(sel[:, 0]) : np.max(sel[:, 0]) + 1,
                        np.min(sel[:, 1]) : np.max(sel[:, 1]) + 1,
                        np.min(sel[:, 2]) : np.max(sel[:, 2]) + 1,
                    ]
                )
            )

    # -----------------------------------------------------------------------------
    # plotting: create a new figure, plot into it, then close it so it never gets
    # displayed

    # turn interactive plotting off unless interactive
    if not INTERACTIVE:
        plt.ioff()

    # compute layout
    if LAYOUT == "default":
        myLayout = computeLayout(len(CutsRRAS))
    else:
        myLayout = LAYOUT

    # create list of de-facto layouts
    myLayoutList = list()
    for axsx in range(myLayout[0]):
        for axsy in range(myLayout[1]):
            myLayoutList.append((axsx, axsy))

    # create subplots
    fig, axs = plt.subplots(myLayout[0], myLayout[1])
    axs = np.reshape(axs, myLayout)

    # adjust layout
    fig.set_size_inches([FIGSIZE * myLayout[1], FIGSIZE * myLayout[0]])
    fig.set_dpi(FIGDPI)
    fig.set_facecolor("black")
    fig.set_tight_layout({"pad": 0})
    fig.subplots_adjust(wspace=0)

    # -----------------------------------------------------------------------------
    # plot each panel

    for p in range(len(CutsRRAS)):
        logging.info("Panel " + str(p))

        axsx = myLayoutList[p][0]
        axsy = myLayoutList[p][1]

        # determine dimensions
        if CutsRRAS[p][0] == "x":
            # x axis of the image should be towards anterior, y axis should be towards superior in RAS image
            dims = (1, 2)
            # determine extent
            extent = (
                rasIdxFlat3[0, dims[0]],
                rasIdxFlat3[-1, dims[0]],
                rasIdxFlat3[0, dims[1]],
                rasIdxFlat3[-1, dims[1]],
            )
            # imshow puts the first dimension (rows) of the data on the y axis, and the second (columns) on the x axis
            cor = np.where(m[dims[0], 0:3])[0]
            axi = np.where(m[dims[1], 0:3])[0]
            #
            if axi < cor:
                axs[axsx, axsy].imshow(
                    normValsRAS[p], cmap="gray", origin="lower", extent=extent
                )
                if aseg is not None:
                    axs[axsx, axsy].imshow(
                        asegValsRAS[p] + 0.5,
                        cmap=lutMap,
                        origin="lower",
                        extent=extent,
                        vmin=0,
                        vmax=len(lutTab),
                        alpha=ALPHA,
                    )
            else:
                axs[axsx, axsy].imshow(
                    normValsRAS[p].transpose(),
                    cmap="gray",
                    origin="lower",
                    extent=extent,
                )
                if aseg is not None:
                    axs[axsx, axsy].imshow(
                        asegValsRAS[p].transpose() + 0.5,
                        cmap=lutMap,
                        origin="lower",
                        extent=extent,
                        vmin=0,
                        vmax=len(lutTab),
                        alpha=ALPHA,
                    )
        elif CutsRRAS[p][0] == "y":
            # x axis of the image should be towards right, y axis should be towards superior in RAS image
            dims = (0, 2)
            # determine extent
            extent = (
                rasIdxFlat3[0, dims[0]],
                rasIdxFlat3[-1, dims[0]],
                rasIdxFlat3[0, dims[1]],
                rasIdxFlat3[-1, dims[1]],
            )
            # imshow puts the first dimension (rows) of the data on the y axis, and the second (columns) on the x axis
            sag = np.where(m[dims[0], 0:3])[0]
            axi = np.where(m[dims[1], 0:3])[0]
            #
            if axi < sag:
                axs[axsx, axsy].imshow(
                    normValsRAS[p], cmap="gray", origin="lower", extent=extent
                )
                if aseg is not None:
                    axs[axsx, axsy].imshow(
                        asegValsRAS[p] + 0.5,
                        cmap=lutMap,
                        origin="lower",
                        extent=extent,
                        vmin=0,
                        vmax=len(lutTab),
                        alpha=ALPHA,
                    )
            else:
                axs[axsx, axsy].imshow(
                    normValsRAS[p].transpose(),
                    cmap="gray",
                    origin="lower",
                    extent=extent,
                )
                if aseg is not None:
                    axs[axsx, axsy].imshow(
                        asegValsRAS[p].transpose() + 0.5,
                        cmap=lutMap,
                        origin="lower",
                        extent=extent,
                        vmin=0,
                        vmax=len(lutTab),
                        alpha=ALPHA,
                    )
        elif CutsRRAS[p][0] == "z":
            # x axis of the image should be towards right, y axis should be towards anterior in RAS image
            dims = (0, 1)
            # determine extent
            extent = (
                rasIdxFlat3[0, dims[0]],
                rasIdxFlat3[-1, dims[0]],
                rasIdxFlat3[0, dims[1]],
                rasIdxFlat3[-1, dims[1]],
            )
            # imshow puts the first dimension (rows) of the data on the y axis, and the second (columns) on the x axis
            sag = np.where(m[dims[0], 0:3])[0]
            cor = np.where(m[dims[1], 0:3])[0]
            if cor < sag:
                axs[axsx, axsy].imshow(
                    normValsRAS[p], cmap="gray", origin="lower", extent=extent
                )
                if aseg is not None:
                    axs[axsx, axsy].imshow(
                        asegValsRAS[p] + 0.5,
                        cmap=lutMap,
                        origin="lower",
                        extent=extent,
                        vmin=0,
                        vmax=len(lutTab),
                        alpha=ALPHA,
                    )
            else:
                axs[axsx, axsy].imshow(
                    normValsRAS[p].transpose(),
                    cmap="gray",
                    origin="lower",
                    extent=extent,
                )
                if aseg is not None:
                    axs[axsx, axsy].imshow(
                        asegValsRAS[p].transpose() + 0.5,
                        cmap=lutMap,
                        origin="lower",
                        extent=extent,
                        vmin=0,
                        vmax=len(lutTab),
                        alpha=ALPHA,
                    )

        # prepare plot
        if rasIdxFlat3[0, dims[0]] > rasIdxFlat3[-1, dims[0]]:
            axs[axsx, axsy].invert_xaxis()
        if rasIdxFlat3[0, dims[1]] > rasIdxFlat3[-1, dims[1]]:
            axs[axsx, axsy].invert_yaxis()

        axs[axsx, axsy].set_axis_off()
        axs[axsx, axsy].set_aspect("equal")

        if XLIM is not None:
            axs[axsx, axsy].set_xlim(XLIM[p])

        if YLIM is not None:
            axs[axsx, axsy].set_ylim(YLIM[p])

        # determine left-right orientation for coronal and axial views
        if ORIENTATION == "radiological":
            if CutsRRAS[p][0] == "y" or CutsRRAS[p][0] == "z":
                axs[axsx, axsy].invert_xaxis()

        # now plot
        for s in range(len(surf)):
            if len(LVL[s][p][0][0]) > 0:
                logging.info("Surface " + str(s))

                # create array of line segments
                tmpx = list()
                tmpy = list()

                for i in range(len(LVL[s][p][1][0])):
                    tmpx.append(
                        (
                            LVL[s][p][0][0][LVL[s][p][1][0][i][0] - 1][dims[0]],
                            LVL[s][p][0][0][LVL[s][p][1][0][i][1] - 1][dims[0]],
                        )
                    )
                    tmpy.append(
                        (
                            LVL[s][p][0][0][LVL[s][p][1][0][i][0] - 1][dims[1]],
                            LVL[s][p][0][0][LVL[s][p][1][0][i][1] - 1][dims[1]],
                        )
                    )

                tmpx = np.array(tmpx)
                tmpy = np.array(tmpy)

                # remove duplicate points
                tmpxy = np.unique(np.concatenate((tmpx, tmpy), axis=1), axis=0)
                tmpx = tmpxy[:, 0:2]
                tmpy = tmpxy[:, 2:4]

                # remove segments which are de-facto points
                tmpIdx = np.logical_or(
                    np.abs(tmpx[:, 0] - tmpx[:, 1]) > tol,
                    np.abs(tmpy[:, 0] - tmpy[:, 1]) > tol,
                )
                tmpx = tmpx[tmpIdx, :]
                tmpy = tmpy[tmpIdx, :]

                # need to order array of line segments; whenever we encounter a
                # closed loop, we will already plot; otherwise, plot in the end
                sortIdx = np.array(range(0, len(tmpx)))

                tmpxSort = np.array(tmpx[sortIdx[0],], ndmin=2)
                tmpySort = np.array(tmpy[sortIdx[0],], ndmin=2)

                sortIdx = np.delete(sortIdx, sortIdx[0])

                while len(sortIdx) > 1:
                    findIdx = np.array(
                        np.where(
                            np.logical_and(
                                np.abs(
                                    tmpx[sortIdx,] - tmpxSort[tmpxSort.shape[0] - 1, 1]
                                )
                                < tol,
                                np.abs(
                                    tmpy[sortIdx,] - tmpySort[tmpySort.shape[0] - 1, 1]
                                )
                                < tol,
                            )
                        ),
                        ndmin=2,
                    ).T

                    # delete existing finds
                    findIdxKeep = list()
                    for k in range(findIdx.shape[0]):
                        if not np.any(
                            np.all(
                                np.logical_or(
                                    tmpx[sortIdx[findIdx[k, 0]], 0] == tmpxSort,
                                    tmpx[sortIdx[findIdx[k, 0]], 1] == tmpxSort,
                                ),
                                axis=1,
                            )
                        ):
                            findIdxKeep.append(k)
                    findIdx = findIdx[findIdxKeep,]

                    if findIdx.shape[0] == 0:
                        # close loop and plot already
                        axs[axsx, axsy].plot(
                            tmpxSort,
                            tmpySort,
                            color=surfcolor[s],
                            linewidth=np.round(FIGSIZE / 8),
                        )
                        # reset (start new loop)
                        tmpxSort = np.array(tmpx[sortIdx[0],], ndmin=2)
                        tmpySort = np.array(tmpy[sortIdx[0],], ndmin=2)
                        sortIdx = np.delete(sortIdx, 0)
                    elif findIdx.shape[0] == 1:
                        # add to current set
                        if findIdx[0, 1] == 0:
                            tmpxSort = np.append(
                                tmpxSort,
                                np.array(tmpx[sortIdx[findIdx[0, 0]], ::1], ndmin=2),
                                axis=0,
                            )
                            tmpySort = np.append(
                                tmpySort,
                                np.array(tmpy[sortIdx[findIdx[0, 0]], ::1], ndmin=2),
                                axis=0,
                            )
                        elif findIdx[0, 1] == 1:
                            tmpxSort = np.append(
                                tmpxSort,
                                np.array(tmpx[sortIdx[findIdx[0, 0]], ::-1], ndmin=2),
                                axis=0,
                            )
                            tmpySort = np.append(
                                tmpySort,
                                np.array(tmpy[sortIdx[findIdx[0, 0]], ::-1], ndmin=2),
                                axis=0,
                            )
                        sortIdx = np.delete(sortIdx, findIdx[0, 0])
                    elif findIdx.shape[0] > 1:
                        warnings.warn(
                            "WARNING: a problem occurred with the surface overlays",
                            stacklevel = 2
                        )
                # now final plot
                axs[axsx, axsy].plot(
                    tmpxSort,
                    tmpySort,
                    color=surfcolor[s],
                    linewidth=np.round(FIGSIZE / 8),
                )

    # -----------------------------------------------------------------------------
    # output

    if not INTERACTIVE:
        plt.savefig(OUTFILE, facecolor=fig.get_facecolor())
        plt.close(fig)

    # -----------------------------------------------------------------------------
    #



"""Handle optional dependency imports.

Inspired from pandas: https://pandas.pydata.org/
"""

import importlib

# A mapping from import name to package name (on PyPI) when the package name
# is different.
INSTALL_MAPPING = {
    "sksparse": "scikit-sparse",
}


def import_optional_dependency(
    name: str,
    extra: str = "",
    raise_error: bool = True,
):
    """Import an optional dependency.

    By default, if a dependency is missing an ImportError with a nice message
    will be raised.

    Parameters
    ----------
    name : str
        The module name.
    extra : str, default=""
        Additional text to include in the ImportError message.
    raise_error : bool, default=True
        What to do when a dependency is not found.
        * True : Raise an ImportError.
        * False: Return None.

    Returns
    -------
    module : Optional[ModuleType]
        The imported module when found.
        None is returned when the package is not found and raise_error is
        False.

    Raises
    -------
    ImportError
        dependency not found; see raise_error
    """

    package_name = INSTALL_MAPPING.get(name)
    install_name = package_name if package_name is not None else name

    try:
        module = importlib.import_module(name)
    except ImportError as err:
        if raise_error:
            raise ImportError(
                f"Missing optional dependency '{install_name}'. {extra} "
                f"Use pip or conda to install {install_name}."
            ) from err
        else:
            return None

    return module



"""
This module provides a function to check the topology of left and right surfaces

"""

# -----------------------------------------------------------------------------


def checkTopology(subjects_dir, subject):
    """
    Check the topology of left and right surfaces.

    This function extracts information about the number of holes and defects
    in the left and right hemisphere, and also returns topological fixing time.

    Parameters
    ----------
    subjects_dir : str
        The directory containing subject data.
    subject : str
        The name of the subject.

    Returns
    -------
    lh_holes : int
        Number of holes in the left hemisphere.
    rh_holes : int
        Number of holes in the right hemisphere.
    lh_defects : int
        Number of defects in the left hemisphere.
    rh_defects : int
        Number of defects in the right hemisphere.
    topo_time_lh : float
        Topological fixing time for the left hemisphere.
    topo_time_rh : float
        Topological fixing time for the right hemisphere.

    Notes
    -----
    Requires a valid scripts/recon-all.log file. If not found, NaNs will be
    returned.
    """
    # Imports

    import logging
    import os
    import warnings

    import numpy as np
    from lapy import TriaMesh

    # Settings

    logging.captureWarnings(True)

    # Message

    logging.info("Checking topology of the surfaces ...")

    # Get the logfile, and return with NaNs if unsuccessful:

    path_log_file = os.path.join(subjects_dir, subject, "scripts", "recon-all.log")

    if os.path.exists(path_log_file):
        with open(path_log_file) as logfile:
            lines_log_file = logfile.read().splitlines()
    else:
        warnings.warn("WARNING: could not find " + path_log_file + ", returning NaNs.",
            stacklevel = 2)
        return np.nan, np.nan, np.nan, np.nan, np.nan, np.nan

    # Initialize

    lh_holes = np.nan
    rh_holes = np.nan
    lh_defects = np.nan
    rh_defects = np.nan
    topo_time_lh = np.nan
    topo_time_rh = np.nan

    # Extract info from surfaces

    lh_orig_nofix = os.path.join(subjects_dir, subject, "surf", "lh.orig.nofix")
    if os.path.exists(lh_orig_nofix):
        lh_euler = TriaMesh.read_fssurf(lh_orig_nofix).euler()
        lh_holes = int((2 - lh_euler ) / 2)
        logging.info("Number of holes in the left hemisphere: " + str(lh_holes))
    else:
        logging.info("Could not find: " + lh_orig_nofix)

    rh_orig_nofix = os.path.join(subjects_dir, subject, "surf", "rh.orig.nofix")
    if os.path.exists(rh_orig_nofix):
        rh_euler = TriaMesh.read_fssurf(rh_orig_nofix).euler()
        rh_holes = int((2 - rh_euler ) / 2)
        logging.info("Number of holes in the right hemisphere: " + str(rh_holes))
    else:
        logging.info("Could not find: " + rh_orig_nofix)

    # Extract info from logfile

    foundDefectsLH = False
    foundTopoLH = False

    for line_log_file in lines_log_file:
        # Look for the number of defects
        if "defects found" in line_log_file and foundDefectsLH is False:
            lh_defects = line_log_file.split()[0]
            lh_defects = int(lh_defects)
            logging.info("Number of defects in the left hemisphere: " + str(lh_defects))
            foundDefectsLH = True
        elif "defects found" in line_log_file and foundDefectsLH is True:
            rh_defects = line_log_file.split()[0]
            rh_defects = int(rh_defects)
            logging.info(
                "Number of defects in the right hemisphere: " + str(rh_defects)
            )

        # Look for the topological fixing time in the log file
        if "topology fixing took" in line_log_file and foundTopoLH is False:
            topo_time_lh = line_log_file.split()[3]
            logging.info(
                "Topological fixing time for the left hemisphere: "
                + str(topo_time_lh)
                + " min"
            )
            foundTopoLH = True
        elif "topology fixing took" in line_log_file and foundTopoLH is True:
            topo_time_rh = line_log_file.split()[3]
            logging.info(
                "Topological fixing time for the right hemisphere: "
                + str(topo_time_rh)
                + " min"
            )

    # Return

    return lh_holes, rh_holes, lh_defects, rh_defects, topo_time_lh, topo_time_rh



"""
This module provides a function to check the SNR of the white and gray matter

"""

# -----------------------------------------------------------------------------


def checkSNR(
    subjects_dir,
    subject,
    nb_erode=3,
    ref_image="norm.mgz",
    aparc_image="aparc+aseg.mgz",
):
    """
    A function to check the SNR of the white and gray matter.

    This function checks the SNR of the white and the gray matter. The white
    matter segmentation is taken from the aparc+aseg image, and the gray matter
    from the aseg image. The white matter is eroded by three voxels to
    ignore partial volumes. For the gray matter, this is not possible because
    the layer is already very thin. An erosion would eliminate nearly the whole
    signal.

    Parameters
    ----------
    subjects_dir : str
        The directory containing subject data.
    subject : str
        The name of the subject.
    nb_erode : int, optional
        The number of erosions, default is 3.
    ref_image : str, optional
        The reference image, default is "norm.mgz", can be changed to "orig.mgz".
    aparc_image : str, optional
        The aparc+aseg image, default is "aparc+aseg.mgz", can
        be changed to "aparc+aseg.orig.mgz" for FastSurfer output.

    Returns
    -------
    wm_snr : float
        The signal-to-noise ratio of the white matter.
    gm_snr : float
        The signal-to-noise ratio of the gray matter.

    Notes
    -----
    It requires valid mri/norm.mgz, mri/aseg.mgz, and mri/aparc+aseg.mgz files for
    FreeSurfer output, and valid mri/norm.mgz, mri/aseg.mgz, and
    mri/aparc+aseg.orig.mgz files for FastSurfer output.
    If not found, NaNs will be returned.
    """
    # Imports

    import logging
    import os
    import warnings

    import nibabel as nib
    import numpy as np
    from skimage.morphology import binary_erosion

    # Settings

    logging.captureWarnings(True)

    # Message

    logging.info("Computing white and gray matter SNR for " + ref_image + " ...")

    # Get data

    path_reference_image = os.path.join(subjects_dir, subject, "mri", ref_image)
    if os.path.exists(path_reference_image):
        norm = nib.load(path_reference_image)
        norm_data = norm.get_fdata()
    else:
        warnings.warn(
            "WARNING: could not open " + path_reference_image + ", returning NaNs.",
            stacklevel = 2
        )
        return np.nan, np.nan

    path_aseg = os.path.join(subjects_dir, subject, "mri", "aseg.mgz")
    if os.path.exists(path_aseg):
        aseg = nib.load(path_aseg)
        data_aseg = aseg.get_fdata()
    else:
        warnings.warn("WARNING: could not open " + path_aseg + ", returning NaNs.",
            stacklevel = 2)
        return np.nan, np.nan

    path_aparc_aseg = os.path.join(subjects_dir, subject, "mri", aparc_image)
    if os.path.exists(path_aparc_aseg):
        inseg = nib.load(path_aparc_aseg)
        data_aparc_aseg = inseg.get_fdata()
    else:
        warnings.warn(
            "WARNING: could not open " + path_aparc_aseg + ", returning NaNs.",
            stacklevel = 2
        )
        return np.nan, np.nan

    # Process white matter image

    # Create 3D binary data where the white matter locations are encoded with 1, all the others with zero
    b_wm_data = np.zeros(norm.shape)

    # The following keys represent the white matter labels in the aparc+aseg image
    wm_labels = [2, 41, 7, 46, 251, 252, 253, 254, 255, 77, 78, 79]

    # Find the wm labels in the aparc+aseg image and set the locations in the binary image to one
    for i in wm_labels:
        x, y, z = np.where(data_aparc_aseg == i)
        b_wm_data[x, y, z] = 1

    # Erode white matter image
    nb_erode = nb_erode
    b_wm_data = binary_erosion(b_wm_data, np.ones((nb_erode, nb_erode, nb_erode)))

    # Computation of the SNR of the white matter
    x, y, z = np.where(b_wm_data == 1)
    signal_wm = norm_data[x, y, z]
    signal_wm_mean = np.mean(signal_wm)
    signal_wm_std = np.std(signal_wm)
    wm_snr = signal_wm_mean / signal_wm_std
    logging.info("White matter signal to noise ratio: " + f"{wm_snr:.4}")

    # Process gray matter image

    # Create 3D binary data where the gray matter locations are encoded with 1, all the others with zero
    b_gm_data = np.zeros(norm.shape)

    # The following keys represent the gray matter labels in the aseg image
    gm_labels = [3, 42]

    # Find the gm labels in the aseg image and set the locations in the binary image to one
    for i in gm_labels:
        x, y, z = np.where(data_aseg == i)
        b_gm_data[x, y, z] = 1

    # Computation of the SNR of the gray matter
    x, y, z = np.where(b_gm_data == 1)
    signal_gm = norm_data[x, y, z]
    signal_gm_mean = np.mean(signal_gm)
    signal_gm_std = np.std(signal_gm)
    gm_snr = signal_gm_mean / signal_gm_std
    logging.info("Gray matter signal to noise ratio: " + f"{gm_snr:.4}")

    # Return
    return wm_snr, gm_snr



"""
This module provides a function to create surface plots
"""
# -----------------------------------------------------------------------------


def createSurfacePlots(SUBJECT, SUBJECTS_DIR, SURFACES_OUTDIR, VIEWS, FASTSURFER):
    """
    Create surface plots.

    Parameters
    ----------
    SUBJECT : str
        The subject.
    SUBJECTS_DIR : str
        The subjects directory.
    SURFACES_OUTDIR : str
        The output directory for surface plots.
    VIEWS : list
        List of views for which surface plots should be created.
    FASTSURFER : bool
        Flag indicating whether FastSurfer processing was used.

    Returns
    -------
    None
        The function returns nothing.
    """
    # imports

    import os

    from whippersnappy import core
    from whippersnappy.types import ViewType

    # -----------------------------------------------------------------------------
    # import surfaces and overlays

    triaPialL = os.path.join(SUBJECTS_DIR, SUBJECT, "surf", "lh.pial")
    triaPialR = os.path.join(SUBJECTS_DIR, SUBJECT, "surf", "rh.pial")
    triaInflL = os.path.join(SUBJECTS_DIR, SUBJECT, "surf", "lh.inflated")
    triaInflR = os.path.join(SUBJECTS_DIR, SUBJECT, "surf", "rh.inflated")

    if FASTSURFER is True:
        annotL = os.path.join(SUBJECTS_DIR, SUBJECT, "label", "lh.aparc.DKTatlas.annot")
        annotR = os.path.join(SUBJECTS_DIR, SUBJECT, "label", "rh.aparc.DKTatlas.annot")
    else:
        annotL = os.path.join(SUBJECTS_DIR, SUBJECT, "label", "lh.aparc.annot")
        annotR = os.path.join(SUBJECTS_DIR, SUBJECT, "label", "rh.aparc.annot")

    # -----------------------------------------------------------------------------
    # plots

    _views_available = [
        'left',
        'right',
        'posterior',
        'anterior',
        'inferior',
        'superior',
    ]

    for view in _views_available:

        fpath_lp = os.path.join(SURFACES_OUTDIR, f"lh.pial.{view}.png")
        fpath_rp = os.path.join(SURFACES_OUTDIR, f"rh.pial.{view}.png")
        fpath_li = os.path.join(SURFACES_OUTDIR, f"lh.inflated.{view}.png")
        fpath_ri = os.path.join(SURFACES_OUTDIR, f"rh.inflated.{view}.png")

        if view in VIEWS:

            if view == "superior":
                wview = ViewType.TOP
            elif view == "inferior":
                wview = ViewType.BOTTOM
            elif view == "anterior":
                wview = ViewType.FRONT
            elif view == "posterior":
                wview = ViewType.BACK
            elif view == "left":
                wview = ViewType.LEFT
            elif view == "right":
                wview = ViewType.RIGHT

            core.snap1(
                meshpath=triaPialL,
                annotpath=annotL,
                outpath=fpath_lp,
                view=wview,
                specular=False
            )
            core.snap1(
                meshpath=triaPialR,
                annotpath=annotR,
                outpath=fpath_rp,
                view=wview,
                specular=False
            )
            core.snap1(
                meshpath=triaInflL,
                annotpath=annotL,
                outpath=fpath_li,
                view=wview,
                specular=False
            )
            core.snap1(
                meshpath=triaInflR,
                annotpath=annotR,
                outpath=fpath_ri,
                view=wview,
                specular=False
            )

        else:
            # remove images potentially created in earlier run but not updated now
            for fpath in [fpath_lp, fpath_rp, fpath_li, fpath_ri]:
                if os.path.isfile(fpath):
                    os.remove(fpath)

    return





# https://setuptools.pypa.io/en/latest/userguide/pyproject_config.html
# If compatibility with legacy builds or versions of tools that don’t support
# certain packaging standards (e.g. PEP 517 or PEP 660), a simple setup.py
# script can be added to your project [1] (while keeping the configuration in
# pyproject.toml):

from setuptools import setup

setup()



import platform
import re
import sys
from functools import partial
from importlib.metadata import requires, version
from typing import IO, Callable, Optional

import psutil


def sys_info(fid: Optional[IO] = None, developer: bool = False):
    """Print the system information for debugging.

    Parameters
    ----------
    fid : file-like, default=None
        The file to write to, passed to :func:`print`.
        Can be None to use :data:`sys.stdout`.
    developer : bool, default=False
        If True, display information about optional dependencies.
    """

    ljust = 26
    out = partial(print, end="", file=fid)
    package = __package__.split(".")[0]

    # OS information - requires python 3.8 or above
    out("Platform:".ljust(ljust) + platform.platform() + "\n")
    # Python information
    out("Python:".ljust(ljust) + sys.version.replace("\n", " ") + "\n")
    out("Executable:".ljust(ljust) + sys.executable + "\n")
    # CPU information
    out("CPU:".ljust(ljust) + platform.processor() + "\n")
    out("Physical cores:".ljust(ljust) + str(psutil.cpu_count(False)) + "\n")
    out("Logical cores:".ljust(ljust) + str(psutil.cpu_count(True)) + "\n")
    # Memory information
    out("RAM:".ljust(ljust))
    out(f"{psutil.virtual_memory().total / float(2 ** 30):0.1f} GB\n")
    out("SWAP:".ljust(ljust))
    out(f"{psutil.swap_memory().total / float(2 ** 30):0.1f} GB\n")

    # dependencies
    out("\nDependencies info\n")
    out(f"{package}:".ljust(ljust) + version(package) + "\n")
    dependencies = [
        elt.split(";")[0].rstrip() for elt in requires(package) if "extra" not in elt
    ]
    _list_dependencies_info(out, ljust, dependencies)

    # extras
    if developer:
        keys = (
            "build",
            "chol",
            "doc",
            "test",
            "style",
        )
        for key in keys:
            dependencies = [
                elt.split(";")[0].rstrip()
                for elt in requires(package)
                if f"extra == '{key}'" in elt or f'extra == "{key}"' in elt
            ]
            if len(dependencies) == 0:
                continue
            out(f"\nOptional '{key}' info\n")
            _list_dependencies_info(out, ljust, dependencies)


def _list_dependencies_info(out: Callable, ljust: int, dependencies: list[str]):
    """List dependencies names and versions.

    Parameters
    ----------
    out : Callable
        output function
    ljust : int
         length of returned string
    dependencies : list[str]
        list of dependencies

    """

    for dep in dependencies:
        # handle dependencies with version specifiers
        specifiers_pattern = r"(~=|==|!=|<=|>=|<|>|===)"
        specifiers = re.findall(specifiers_pattern, dep)
        if len(specifiers) != 0:
            dep, _ = dep.split(specifiers[0])
            while not dep[-1].isalpha():
                dep = dep[:-1]
        # handle dependencies provided with a [key], e.g. pydocstyle[toml]
        if "[" in dep:
            dep = dep.split("[")[0]
        try:
            version_ = version(dep)
        except Exception:
            version_ = "Not found."

        # handle special dependencies with backends, C dep, ..
        if dep in ("matplotlib", "seaborn") and version_ != "Not found.":
            try:
                from matplotlib import pyplot as plt

                backend = plt.get_backend()
            except Exception:
                backend = "Not found"

            out(f"{dep}:".ljust(ljust) + version_ + f" (backend: {backend})\n")

        else:
            out(f"{dep}:".ljust(ljust) + version_ + "\n")



"""
This module provides a function to evaluate potential missegmentation of the
hippocampus and amygdala
"""
# --------------------------------------------------------------------------


def evaluateHippocampalSegmentation(
    SUBJECT,
    SUBJECTS_DIR,
    OUTPUT_DIR,
    CREATE_SCREENSHOT=True,
    SCREENSHOTS_OUTFILE=None,
    SCREENSHOTS_ORIENTATION="radiological",
    HEMI="lh",
    LABEL="T1.v21",
):
    """
    Evaluate potential missegmentation of the hippocampus and amygdala.

    This script evaluates potential missegmentation of the hippocampus and
    amygdala as created by FreeSurfer 7.1's dedicated module:
    https://surfer.nmr.mgh.harvard.edu/fswiki/HippocampalSubfieldsAndNucleiOfAmygdala

    If the corresponding arguments are set to 'True', the script will also
    create screenshots. Resulting files will be saved to the same directory
    as indicated above.

    Parameters
    ----------
    SUBJECT : str
        The subject identifier.
    SUBJECTS_DIR : str
        The directory containing subject data.
    OUTPUT_DIR : str
        The output directory.
    CREATE_SCREENSHOT : bool, optional, default: True
        Whether to create screenshots.
    SCREENSHOTS_OUTFILE : str or list, optional, default: None
        File or list of files for screenshots.
    SCREENSHOTS_ORIENTATION : str, optional, default: "radiological"
        Orientation for screenshots.
    HEMI : str, optional, default: "lh"
        Hemisphere to evaluate, either 'lh' or 'rh'.
    LABEL : str, optional, default: "T1.v21"
        Label for hippocampal and amygdala segmentation.

    Returns
    -------
    None
        This function returns nothing.

    Notes
    -----
    Required Files:
    - mri/[lr]h.hippoAmygLabels-<LABEL>.FSvoxelSpace.mgz
    - mri/norm.mgz
    """
    # imports

    import logging
    import os

    import nibabel as nb
    import numpy as np
    from scipy import ndimage

    from fsqc.createScreenshots import createScreenshots
    from fsqc.fsqcUtils import binarizeImage

    # --------------------------------------------------------------------------
    # check files

    if not os.path.isfile(os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz")):
        logging.error(
            "ERROR: could not find "
            + os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz")
            + ", not running hippocampus module."
        )

        raise ValueError("File not found")

    if not os.path.isfile(
        os.path.join(
            SUBJECTS_DIR,
            SUBJECT,
            "mri",
            HEMI + ".hippoAmygLabels-" + LABEL + ".FSvoxelSpace.mgz",
        )
    ):
        logging.error(
            "ERROR: could not find "
            + os.path.join(
                SUBJECTS_DIR,
                SUBJECT,
                "mri",
                HEMI + ".hippoAmygLabels-" + LABEL + ".FSvoxelSpace.mgz",
            )
            + ", not running hippocampus module."
        )

        raise ValueError("File not found")

    if SCREENSHOTS_OUTFILE is None:
        SCREENSHOTS_OUTFILE = os.path.join(OUTPUT_DIR, "hippocampus.png")

    # --------------------------------------------------------------------------
    # create mask

    binarizeImage(
        os.path.join(
            SUBJECTS_DIR,
            SUBJECT,
            "mri",
            HEMI + ".hippoAmygLabels-" + LABEL + ".FSvoxelSpace.mgz",
        ),
        os.path.join(OUTPUT_DIR, "hippocampus-" + HEMI + ".mgz"),
        match=None,
    )

    # --------------------------------------------------------------------------
    # get centroids

    seg = nb.load(
        os.path.join(
            SUBJECTS_DIR,
            SUBJECT,
            "mri",
            HEMI + ".hippoAmygLabels-" + LABEL + ".FSvoxelSpace.mgz",
        )
    )
    seg_data = seg.get_fdata()
    seg_labels = np.setdiff1d(np.unique(seg_data), 0)

    centroids = np.array(ndimage.center_of_mass(seg_data, seg_data, seg_labels))
    centroids = np.concatenate((seg_labels[:, np.newaxis], centroids), axis=1)

    vox2ras_tkr = seg.header.get_vox2ras_tkr()

    ctr_tkr = np.concatenate(
        (centroids[:, 1:4], np.ones((centroids.shape[0], 1))), axis=1
    )
    ctr_tkr = np.matmul(vox2ras_tkr, ctr_tkr.T).T
    ctr_tkr = np.concatenate(
        (np.array(centroids[:, 0], ndmin=2).T, ctr_tkr[:, 0:3]), axis=1
    )

    # [7004, 237, 238]

    ctr_tkr_x0 = ctr_tkr[np.argwhere(ctr_tkr[:, 0] == 237), 1]
    ctr_tkr_y0 = ctr_tkr[np.argwhere(ctr_tkr[:, 0] == 237), 2]
    ctr_tkr_z0 = ctr_tkr[np.argwhere(ctr_tkr[:, 0] == 237), 3]

    # set ranges for cropping the image (assuming RAS coordinates)

    ctr_tkr_x0_xlim = [ctr_tkr_y0 - 40, ctr_tkr_y0 + 40]
    ctr_tkr_y0_xlim = [ctr_tkr_x0 - 20, ctr_tkr_x0 + 20]
    ctr_tkr_z0_xlim = [ctr_tkr_x0 - 20, ctr_tkr_x0 + 20]

    ctr_tkr_x0_ylim = [ctr_tkr_z0 - 40, ctr_tkr_z0 + 40]
    ctr_tkr_y0_ylim = [ctr_tkr_z0 - 40, ctr_tkr_z0 + 40]
    ctr_tkr_z0_ylim = [ctr_tkr_y0 - 40, ctr_tkr_y0 + 40]

    XLIM = [ctr_tkr_x0_xlim, ctr_tkr_y0_xlim, ctr_tkr_z0_xlim]
    YLIM = [ctr_tkr_x0_ylim, ctr_tkr_y0_ylim, ctr_tkr_z0_ylim]

    # --------------------------------------------------------------------------
    # create screenshots

    if CREATE_SCREENSHOT is True:
        createScreenshots(
            SUBJECT=SUBJECT,
            SUBJECTS_DIR=SUBJECTS_DIR,
            INTERACTIVE=False,
            VIEWS=[("x", ctr_tkr_x0), ("y", ctr_tkr_y0), ("z", ctr_tkr_z0)],
            LAYOUT=(1, 3),
            BASE=os.path.join(SUBJECTS_DIR, SUBJECT, "mri", "norm.mgz"),
            OVERLAY=os.path.join(
                SUBJECTS_DIR,
                SUBJECT,
                "mri",
                HEMI + ".hippoAmygLabels-" + LABEL + ".FSvoxelSpace.mgz",
            ),
            SURF=None,
            OUTFILE=SCREENSHOTS_OUTFILE,
            ORIENTATION=SCREENSHOTS_ORIENTATION,
            XLIM=XLIM,
            YLIM=YLIM,
        )
